"""This module handles lists"""

import itertools

from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple
from typing import TypeVar

Unique = TypeVar("Unique")  # Generic type


class Nones(list):
    limit: int = 0

    def __init__(self, limit: int, *args):
        self.limit = limit
        self.check(len(args))
        self.args = args
        super().__init__(*args)

    def check(self, i) -> bool:
        if self.limit < 0:
            return True
        if len(args) <= self.limit:
            return True
        raise TypeError(f"{self.__class__.__name__} has nothing at {i=}")

    def __getitem__(self, i) -> Any:
        self.check(i)
        return super().__getitem__(i)


class One(Nones):
    def __init__(self, *args):
        self.one, *_ = _args
        super().__init__(1, [self.one])


class Two(Nones):
    def __init__(self, *args):
        self.one, self.two, *_ = _args
        super().__init__(2, args)


class Many(Two):
    def __init__(self, *args):
        super().__init__(8, args)


class Lots(Many):
    def __init__(self, *args):
        self.limit = -1
        super().__init__(-1, args)


# https://stackoverflow.com/a/25464724/500942
class Uniques(list):
    """A list of unique items

    Uniqueness is checked when appending to the list
    """

    def __init__(self, items=None):
        super().__init__([])
        self.extend(items or [])

    def predicate(self, _: Unique) -> bool:
        """Subclasses can exclude items which return False"""
        return True

    def convert(self, item: Unique) -> Unique:
        return item

    def append(self, item: Unique) -> None:
        if item in self:
            return
        if not self.predicate(item):
            return
        super().append(self.convert(item))

    def extend(self, items: Iterable[Unique]) -> None:
        for item in items or []:
            self.append(item)


class UniquelyTrues(Uniques):
    """A unique list of items all being true-ish"""

    def predicate(self, item: Unique) -> bool:
        return bool(item)


def de_duplicate(items: List) -> List:
    """Remove any duplicate item, preserving order

    >>> assert de_duplicate([1, 9, 2, 8, 1, 7, 2]) == [1, 9, 2, 8, 7]
    """
    return list(Uniques(items))


def flatten(list_of_lists: List[List]) -> List:
    """Reduce a lists of lists to a list, comprehensively

    >>> assert flatten([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to Alex Martelli at
        https://stackoverflow.com/a/952952/500942
    """
    return [item for list_ in list_of_lists for item in list_]


def flatten_(list_of_lists: List[List]) -> List:
    """Reduce a lists of lists to a list, functionally

    >>> assert flatten_([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to CTT at
        https://stackoverflow.com/a/716482/500942
    """
    return list(itertools.chain.from_iterable(list_of_lists))


def as_list(item):
    """Make item a list from types which can index, have items, or can pop"""
    try:
        item.index
        return list(item)
    except AttributeError:
        try:
            return list(item.items())
        except AttributeError:
            try:
                item.pop
                return list(item)
            except AttributeError:
                raise TypeError(f"Cannot make a list from {item!r}")


def splits(predicate: Callable, items: List) -> Tuple[List[Any], List[Any]]:
    """Split a list into 2 lists based one the predicate

    >>> is_odd = lambda x: bool(x % 2)
    >>> odds, evens = splits(is_odd, [0, 1, 2, 3, 4, 5, 6])
    >>> assert odds == [1, 3, 5]
    >>> assert evens == [0, 2, 4, 6]
    """
    one: List[Any] = []
    two: List[Any] = []
    for item in items:
        destination = one if predicate(item) else two
        destination.append(item)
    return one, two



"""Show some lines from files

Usage: python3 -m pysyte.kat [options] [files]

"""

from pysyte.cli import arguments
from pysyte.cli import lines
from pysyte.cli.app import App


def parse_args():
    parser = arguments.parser(__doc__)
    line_parser = lines.add_args(parser)
    return line_parser.parse_args()


def kat(app):
    """Run kat"""
    for file in app.args.files():
        text = file.read()
        start, lines_in = arg_lines(text, app.args)
        lines_out = args.sed(lines_in, start)
        text_ = "\n".join(lines_out)
        print(f"{text_}\n")
    return True


def main():
    args = parse_args()
    with App(kat, args) as app:
        app.run(app.args)


if __name__ == "__main__":
    main()



"""Run scripts from the command_line

This module was a simplifying proxy to stdlib's sys.exit()
    but it's grown since then
"""

import bdb
import sys
from dataclasses import dataclass
from typing import Optional

from pysyte.cli import arguments
from pysyte.cli import app
from pysyte.cli.config import load_configs
from pysyte.types.paths import makepath
from pysyte.types.methods import Callable
from pysyte.types.methods import Method


class MainMethod(Method):
    def __init__(self, method):
        super().__init__(method)
        self.in_main_module = self.module.__name__ == "__main__"
        self.needs_args = self.argcount > 0
        self.needs_one_arg = self.argcount == 1

    def __call__(self, *args_, **kwargs):
        return self.method(*args_, **kwargs)

    @property
    def doc(self) -> str:
        for doc in (super().doc, self.module.__doc__):
            if doc:
                return doc
        return ""


ArgumentsParsers = Callable[[arguments.ArgumentsParser], arguments.ArgumentsParser]


@dataclass
class CallerData:
    method: MainMethod
    add_args: Optional[ArgumentsParsers]


def exit(main: Callable):
    """Exit to shell with error"""
    app.exit(main)


def run(
    main_method: Callable,
    add_args: Optional[ArgumentsParsers] = None,
    post_parse: Optional[Callable] = None,
    usage: Optional[str] = None,
    epilog: Optional[str] = None,
    config_name: Optional[str] = None,
):
    """Run a main_method from command line, parsing arguments

    if add_args(parser) is given it should call parser.add_arg()
        See ArgumentsParser.add_arg(), etc
    if pre_parse() is given then call it instead of sys.argv
    if post_parse(args) is given then it is called with parsed args
        and should return them after any adjustments
        if config_name is given:
            if it is a string then find configs for that
            otherwise use the main_method's script's name
        read config_name, as yaml, to data
        call main_method(args, data)
    else
        call main_method(args)
    """

    class Caller(CallerData):
        def arg_parser(self):
            try:
                parser_ = arguments.parser()
                return self.add_args(parser_)
            except TypeError:
                raise NotImplementedError("Unknown signature for add_args()")

        def parse_args(self):
            if self.add_args:
                parser = self.arg_parser()
                assert parser  # Dont trust callers' coder to return the parser
                return parser.parse_args(post_parser=post_parse)
            assert self.method.needs_one_arg
            return sys.argv[1:]

        def config_name(self, name):
            if not isinstance(name, str):
                return makepath(self.method).name
            p = makepath(name)
            return p.name if p else name

        def config(self):
            return load_configs(self.config_name(config_name))

        def main(self, argument_handler):
            if self.method.needs_args:
                self.args = self.parse_args()
            if config_name:
                return self.method(self.args, self.config())
            if self.method.needs_args:
                return self.method(self.args)
            return self.method()

    caller = Caller(MainMethod(main_method), add_args)
    if caller.method.in_main_module:
        handler = arguments.ArgumentHandler()
        app.exit(lambda: handler.run(caller))



"""Handle ANSI colours for pysyte

Colour sequences are derived from the table at

http://en.wikipedia.org/wiki/ANSI_escape_sequences#Colors

"""
light_colour = "\033[1;%sm"
light_black = light_colour % 30
light_red = light_colour % 31
light_green = light_colour % 32
yellow = light_colour % 33
light_blue = light_colour % 34
light_magenta = light_colour % 35
light_cyan = light_colour % 36
white = light_colour % 37
dark_colour = "\033[0;%sm"
black = light_colour % 30
red = dark_colour % 31
green = dark_colour % 32
brown = dark_colour % 33
blue = dark_colour % 34
magenta = dark_colour % 35
cyan = dark_colour % 36
grey = dark_colour % 37
off = "\033[0m"


def highlighted(colour, text):
    return f"{colour}{text}{off}"



"""Classes to proxy other objects"""

from typing import Any


class Proxy(object):
    """Offers attributes from self, or another object"""

    def __init__(self, value: Any, name: str = "", **kwargs: dict) -> None:
        super().__init__()
        self.__dict__["_proxy"] = value
        if name:
            self.__dict__[name] = value

    def __getattr__(self, name: str) -> Any:
        try:
            return getattr(self._proxy, name)
        except AttributeError:
            raise AttributeError(
                f"'{self.__class__.__name__}' has no attribute '{name}'"
            )

    def __setattr__(self, name: str, value: Any) -> None:
        proxy = self.__dict__["_proxy"]
        try:
            getattr(proxy, name)
            setattr(proxy, name, value)
        except AttributeError:
            self.__dict__[name] = value




"""This module provides some known ANSI escape codes which work with terminals

All numbers used in this file are values usable in an ANSI escape sequence
    See http://en.wikipedia.org/wiki/ANSI_escape_code for values
"""


def escape(string):
    return f"\033{string}"


def escape_sequence(string):
    return escape(f"[{string}m")


def bold():
    return escape_sequence("1")


def no_bold():
    return escape_sequence("22")


def no_colour():
    return escape_sequence("2") + escape_sequence("0")


def no_prompt():
    return prompt(no_colour())


def _colour_16(ground, i):
    if i > 7:
        prefix = bold()
        i = i - 8
    else:
        prefix = ""
    escaped = escape(f"[{ground}{i}m")
    return f"{prefix}{escaped}"


def _colour_256(ground, i):
    return escape_sequence(f"{ground};5;{i}")


def _background_16(i):
    return _colour_16(4, i)


def _background_256(i):
    return _colour_256(48, i)


def _foreground_16(i):
    return _colour_16(3, i)


def _foreground_256(i):
    return _colour_256(38, i)


def _small_colour_number(i):
    return i < 16


def foreground(i):
    if not i:
        return ""
    return _small_colour_number(i) and _foreground_16(i) or _foreground_256(i)


def background(i):
    if not i:
        return ""
    return _small_colour_number(i) and _background_16(i) or _background_256(i)


def prompt(string):
    return f"\001{string}\002"


def colour_string(colour, text):
    stop = no_colour() if colour else ""
    return f"{colour}{text}{stop}"


def foreground_string(text, i):
    return colour_string(foreground(i), text)


def background_string(text, i):
    return colour_string(background(i), text)


def grounds_string(text, background_colour, foreground_colour):
    return "%s%s%s%s" % (
        background(background_colour),
        foreground(foreground_colour),
        text,
        no_colour(),
    )


def prompt_string(text, i):
    string = foreground(i)
    return f"{prompt(string)}{text}{no_prompt()}"



"""This module handles numbers for pysyte

>>> from pysyte.types import numbers

integers and strings work like int() for both
>>> assert numbers.as_int(5) == 5 == numbers.as_int(" 5  ")
>>> assert numbers.inty(5) == 5 == numbers.inty("5   ")

inty() does extra computation is done for things with a length

>>> assert numbers.inty("lll") == 3
>>> assert (
...     numbers.inty(
...         [
...             3,
...         ]
...     )
...     == 1
... )

as_int() does not
>>> try:
...     numbers.as_int("lll")
...     assert False
... except numbers.NAN:
...     pass
...
>>> try:
...     numbers.as_int(
...         [
...             3,
...         ]
...     )
...     assert False
... except numbers.NAN:
...     pass
...
"""


class NAN(ValueError):
    def __init__(self, value):
        super().__init__(f"NAN: {value}")


def _to_number(value, kind, default):
    """Convert the value to that kind of int, using that default

    >>> assert _to_number(3, int, 7) == 3
    >>> assert _to_number(3, float, 7) == 3.0
    >>> assert _to_number(None, int, 7) == 7
    >>> assert _to_number(None, float, 7) == 7.0
    >>> assert _to_number(3, int, None) == 3
    >>> assert _to_number(3, float, None) == 3.0
    """
    try:
        return kind(value)
    except (ValueError, TypeError):
        if default is None:
            raise NAN(value)
        if isinstance(value, (type(None),)):
            return kind(default)
        try:
            return len(value)
        except TypeError:
            raise ValueError(value)


as_int = lambda x: _to_number(x, int, None)
as_float = lambda x: _to_number(x, float, None)

inty = lambda x: _to_number(x, int, x)
floaty = lambda x: _to_number(x, float, x)



"""Miscellaneous string handlers"""

from functools import total_ordering


class Repper:
    """Make a nice repr(x) for any x that can do str(x)"""

    def __repr__(self) -> str:
        string = str(self)
        return f'<{self.__class__.__name__} "{string}">'


@total_ordering
class Stringer(Repper):
    """A Mixin to provide more ops to a class that can str()"""

    def __str__(self):
        raise NotImplementedError

    def __hash__(self):
        """Derive a hash value from the string"""
        return hash(str(self))

    def __eq__(self, other):
        """Whether this equals other, compared as strings"""
        return str(self) == str(other)

    def __ne__(self, other):
        """Whether this differs from other, compared as strings"""
        return str(self) != str(other)

    def __lt__(self, other):
        """Whether this is less than other, compared as strings"""
        return str(self) < str(other)



"""Handle colours for screens and walls"""

from pysyte.colours import colour_numbers


class Colour(object):
    def __init__(self, name):
        self._id = colour_numbers.name_to_id(name)
        self._name = name

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name



"""Provide methods to put colours in texts


Inspired by http://stackoverflow.com/questions/384076/how-can-i-make-the-python-logging-output-to-be-colored  # noqa
# Conversion from RGB to xterm 256 colours based on values in http://www.frexx.de/xterm-256-notes/data/xterm-colortest  # noqa
Nicer version: https://jonasjacek.github.io/colors/
Conversion from rgb.txt names to RGB inspired by https://github.com/lilydjwg/winterpy/blob/master/pyexe/gui2term.py  # noqa
"""


from __future__ import absolute_import
from functools import partial

from pysyte.colours import ansi_escapes
from pysyte.colours import colour_names
from pysyte.colours import colour_numbers


class ColouredTail(object):
    """The tail of a string that is being coloured

    Also holds its head
        which is all of the string before the tail

    When treated as a string, gives head + tail
    """

    def __init__(self, head, tail):
        self.head = str(head) if head else ""
        self.tail = str(tail) if tail else ""
        self.add_colour_names()

    def add_colour_names(self):
        for colour_name in colour_names.cga():
            attr_name = colour_name.replace(" ", "_")
            setattr(self, attr_name, partial(self.colour_text, colour_name))
        setattr(self, "reddy", partial(self.colour_text, "light red"))
        setattr(self, "none", partial(self.colour_text, None))

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.head!r}{self.tail!r}>"

    def __str__(self):
        return f"{self.head}{self.tail}"

    def __eq__(self, other):
        return str(self) == str(other)

    def colour_text(self, *args):
        try:
            colour_name, text = args
        except ValueError:
            colour_name = None
            text = args[0]
        new = colour_text(colour_name, text, head="")
        return ColouredTail(str(self), new)

    colour = text = colour_text


def colour_text(colour_name, text, head=None):
    return ColouredTail(
        head or "",
        ansi_escapes.foreground_string(
            str(text),
            colour_numbers.name_to_id(colour_name),
        ),
    )


colour = text = colour_text

for name in colour_names.cga():
    globals()[name] = partial(colour_text, name)
none = partial(colour_text, None)


def colour_initial(colour_name, string):
    return colour(colour_name, string[0]).colour(None, string[1:])


def colour_initials(colour_name, strings):
    return "".join([str(colour_initial(colour_name, s)) for s in strings])


def prompt_text(colour_name, string):
    number = colour_numbers.name_to_id(colour_name)
    return ansi_escapes.prompt_string(string, number)



"""Handle config files"""

import yaml

from pysyte.types.dictionaries import NameSpaces


class Config(NameSpaces):
    def __init__(self, data):
        super().__init__(data)


def load(path_to_config):
    with path_to_config.open() as stream:
        data = yaml.safe_load(stream)
        return Config(data)


def dump(data, path_to_config):
    with path_to_config.open("w") as stream:
        yaml.safe_dump(data, stream, explicit_start=True)



"""handle numbers for pysyte

>>> one = lots([''])
>>> assert one.limit == 1
>>> two = lots(["", ""])
>>> assert two.limit == 2
>>> three = lots([1, "7", 9])
>>> assert three.limit == 3
>>> four = lots([9, 5, 3, 0])
>>> assert four.limit == 4

>>> assert two + one == three

"""
from typing import Any

from dataclasses import dataclass
from dataclasses import field


@dataclass
class Nones(list):
    args: list = field(default_factory=list)

    def __post_init__(self):
        breakpoint()
        self.limit = 0
        self.check_args()

    def check_args(self):
        try:
            self.check(len(self.args))
        except IndexError as e:
            args = self.args
            raise ValueError(f"{self.__class__.__name__}({args=!r}): Too many args")

    def __int__(self):
        return len(self.args)

    def __str__(self):
        return str(int(self))

    def __repr__(self):
        return f'<{self.__class__.__name__} {int(self)} {self.args}>'

    def __add__(self, other):
        return lots(self.args + other.args)

    def __sub__(self, other):
        return lots(self.args[-len(other.args) :])  # + other.args)

    def __eq__(self, other):
        return self.__class__.__name__ == other.__class__.__name__

    def __lt__(self, other):
        return int(self) < int(other)

    def __getitem__(self, i) -> Any:
        self.check(i)
        return super().__getitem__(i)

    @property
    def max(self):
        return 5

    def check(self, i) -> bool:
        if self.limit >= self.max:
            return True
        if len(self.args) > self.limit:
            raise IndexError(f"{self.__class__.__name__} has nothing at {i=}")
        return True
        raise TypeError(f"{self.__class__.__name__} has nothing at {i=}")


class One(Nones):
    def __post_init__(self):
        self.limit = 1
        self.check_args()


class Two(Nones):
    def __post_init__(self):
        self.limit = 2
        self.check_args()


class Many(Two):
    def __post_init__(self):
        self.limit = 4
        self.check_args()


class Lots(Many):
    def __post_init__(self):
        self.limit = self.max + 1
        self.check_args()



"""Any option mentioning "a line" means either a number or a regexp"""

import sys
from collections import defaultdict
from functools import partial
from typing import DefaultDict
from typing import List

from pysyte import __version__
from pysyte.cli.arguments import ArgumentsParser
from pysyte.types import lines as pylines
from pysyte.bash.screen import alt_screen


class LinesParser(ArgumentsParser):
    """Add option sets to the parser"""

    def __init__(self, parser_: ArgumentsParser):
        super().__init__(parser_)
        self.version = __version__
        self.groups: DefaultDict[str, List] = defaultdict(list)

    def add_files(self, name=None, action=None):
        name_ = name if name else "files"
        action_ = action if action else "show"
        self.positional(name_, help=f"{name_} to {action_}")
        return self

    def add_to(self, letter, name, group, type_, help_, default=None):
        letter_ = f'-{letter.lstrip("-")}'
        name_ = f'--{name.lstrip("-")}'
        safe_default = default if default else 0 if str(type_).startswith("int") else ""
        method = getattr(self, type_, f"not a method for {letter},{name}")
        assert callable(method), f"self.{type_}() is {method}"
        if type_ == "boolean":
            arg = method(letter_, name_, help=help_)
        else:
            arg = method(letter_, name_, help=help_, default=safe_default)
        self.groups[group].append(arg)

    def add_lines(self):
        args = {
            "a": ["at", "lines", "inty", "show that line"],
            "c": ["copy", "clipboard", "boolean", "copy text to clipboard"],
            # 'd': ['delete', 'lines', 'string', 'lines to be deleted'],
            "e": ["expression", "ed", "string", "sed expression"],
            "f": ["first", "lines", "inty", "the first line to show", "1"],
            "i": [
                "stdin",
                "stdin",
                "boolean",
                "(aka -) wait for text from stdin",
            ],
            "l": ["last", "lines", "inty", "the last line to show", "0"],
            "n": ["numbers", "lines", "boolean", "show line numbers"],
            "p": [
                "paste",
                "clipboard",
                "boolean",
                "paste text from clipboard",
            ],
            "s": ["substitute", "ed", "string", "sed s-expression"],
            "v": ["remove", "lines", "inty", "remove that line"],
            "V": ["version", "version", "boolean", "show version"],
            "w": ["width", "lines", "integer", "max width of lines", "80"],
        }
        [self.add_to(letter, *args) for letter, args in args.items()]
        return self

    def post_parser(self, args):
        if args.version and self.version:
            sys.stdout.write(f"{sys.argv[0]} version: {self.version}")
            raise SystemExit
        args.sed = partial(pylines.sed, args=args)
        args.alt_screen = alt_screen()
        return args


def add_args(old_parser: ArgumentsParser) -> LinesParser:
    """Create a new parser to handle some lines from given parser"""
    result = LinesParser(old_parser.parser)
    result.add_lines()
    return result


def arg_lines(a, b):
    raise NotImplementedError



"""Handle punctuation for pysyte"""

space = " "

bang = "!"
double_quote = '"'
hash_ = "#"
dollar = "$"
modulo = "%"
ampersand = "&"
apostrophe = single_quote = "'"
open_parens = "("
close_parens = ")"
asterisk = "*"
plus = "+"
comma = ","
minus = "-"
dot = full_stop = period = "."
forward_slash = "/"
zero = "0"
colon = ":"
semi_colon = ";"
less = "<"
equal = "="
greater = ">"
question = "?"
at = "@"
open_bracket = "["
backslash = "\\"
close_bracket = "]"
caret = "^"
underline = "_"
backtick = "`"
open_brace = "{"
pipe = "|"
close_brace = "}"

punctuation = f'!"#$%&{apostrophe})(*+,-./:;<=>?@[\\]^_`{open_brace}|{close_brace}'



"""Provide a getch() method to get single chars from a terminal


Do not wait on carriage returns

Adapted from
    http://code.activestate.com/recipes/134892/
    Without the redundant classes of that code
        but keeping the implementatation

    And reducing down to just the tty version
        No need here for Windows/Carbon
"""

import re
import getpass
import signal
import sys
import tty
from curses import ascii
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple

import termios


class NoKeys(StopIteration):
    """A StopIteration caused by running out of keys"""

    pass


def get_ord():
    """The integer ordinal of the next byte read from sys.stdin"""
    return ord(sys.stdin.read(1))


class TerminalContext(object):
    """Context wrapper to set up termios values"""

    def __init__(self):
        self.fd = None
        self.old_settings = None

    def __enter__(self):
        self.fd = sys.stdin.fileno()
        self.old_settings = termios.tcgetattr(self.fd)
        mode = termios.tcgetattr(self.fd)
        mode[tty.LFLAG] = mode[tty.LFLAG] & ~(termios.ECHO | termios.ICANON)
        termios.tcsetattr(self.fd, termios.TCSAFLUSH, mode)
        return self

    def __exit__(self, typ, value, traceback):
        if typ is not None:
            pass  # exception
        if self.old_settings:
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old_settings)


class Timeout(Exception):
    """A timeout has occurred"""

    pass


def timeout_raiser(_signum, _frame):
    """Raise a timeout exception"""
    raise Timeout()


class TimerContext(object):
    """Context wrapper for a timeout"""

    def __init__(self, seconds):
        self.seconds = seconds
        self.old_handler = None
        self.timed_out = False

    def __enter__(self):
        self.old_handler = signal.signal(signal.SIGALRM, timeout_raiser)
        signal.setitimer(signal.ITIMER_REAL, self.seconds)
        return self

    def __exit__(self, typ_, value, traceback):
        signal.setitimer(signal.ITIMER_REAL, 0)
        if self.old_handler:
            signal.signal(signal.SIGALRM, self.old_handler)
        if isinstance(value, Timeout):
            self.timed_out = True
            return True


_key_cache: List[Tuple[int, ...]] = []


def cache_keys(keys):
    """Allow debugging via PyCharm"""
    d = known_keys()
    known_names = dict(zip(d.values(), d.keys()))
    for k in keys:
        i = (ord(k),) if len(k) == 1 else known_names[k]
        _key_cache.insert(0, i)


def _get_keycodes():
    """Read keypress giving a tuple of key codes

    A 'key code' is the ordinal value of characters read

    For example, pressing 'A' will give (65,)
    """
    try:
        return _key_cache.pop()
    except IndexError:
        pass
    result = []
    terminators = "ABCDFHPQRS~"
    with TerminalContext():
        code = get_ord()
        result.append(code)
        if code == 27:
            with TimerContext(0.1) as timer:
                code = get_ord()
            if not timer.timed_out:
                result.append(code)
                result.append(get_ord())
                if 64 < result[-1] < 69:
                    pass
                elif result[1] == 91:
                    while True:
                        code = get_ord()
                        result.append(code)
                        if chr(code) in terminators:
                            break
    return tuple(result)


class ExtendedKey(Exception):
    """Raised for an unrecognised Extended key code"""

    def __init__(self, codes):
        super(ExtendedKey, self).__init__(f"Too many key codes: {codes!r}")
        self.codes = codes


def getch() -> int:
    """Get a char or raise ExtendedKey from a keyboard"""
    keycodes = _get_keycodes()
    code, *codes = keycodes
    if not codes:
        return code
    raise ExtendedKey(keycodes)


def get_codes():
    """Get a tuple of stringified keycodes from the keyboard"""
    return tuple(str(_) for _ in _get_keycodes())


def get_key():
    """Get a key from the keyboard as a string

    A 'key' will be a single char, or the name of an extended key
    """
    character_name = chr
    codes = _get_keycodes()
    if len(codes) == 1:
        code = codes[0]
        if code >= 32:
            return character_name(code)
        return control_key_name(code)
    return get_extended_key_name(codes)


def get_menu(**kwargs):
    key = get_key()
    for name, string in kwargs.items():
        match = re.match(f"^{string}$", key)
        if match:
            return name
        if key in name or key in string:
            return name
    return key


def get_ascii():
    """Get ASCII key from stdin

    return None for others
    """
    try:
        return getch()
    except ExtendedKey:
        return None


def get_ASCII():
    """Get ASCII key from stdin

    raise error on other
    """
    try:
        return getch()
    except ExtendedKey:
        raise KeyboardInterrupt


def get_as_key():
    """Get key from stdin

    return ASCII keys as (single char) strings
    others as tuples
    """
    try:
        return getch()
    except ExtendedKey as e:
        return e.codes


def control_key_name(code):
    """Prefix the name of a control key with '^'"""
    name = chr(code - 1 + ord("A"))
    return f"^{name}"


def get_extended_key_name(codes):
    """Get a name for the extended key with those codes

    Names are defined herein
    """
    return known_keys()[codes]


def known_keys() -> Dict[Tuple[int, ...], str]:
    result = {
        (27, 79, 70): "end",
        (27, 79, 72): "home",
        (27, 79, 80): "F1",
        (27, 79, 81): "F2",
        (27, 79, 82): "F3",
        (27, 79, 83): "F4",
        (27, 91, 49, 53, 126): "F5",
        (27, 91, 49, 53, 59, 50, 126): "F17",
        (27, 91, 49, 55, 126): "F6",
        (27, 91, 49, 55, 59, 50, 126): "F18",
        (27, 91, 49, 56, 126): "F7",
        (27, 91, 49, 56, 59, 50, 126): "F19",
        (27, 91, 49, 57, 126): "F8",
        (27, 91, 49, 59, 50, 65): "&up",
        (27, 91, 49, 59, 50, 66): "&down",
        (27, 91, 49, 59, 50, 67): "&right",
        (27, 91, 49, 59, 50, 68): "&left",
        (27, 91, 49, 59, 50, 80): "F13",
        (27, 91, 49, 59, 50, 83): "F16",
        (27, 91, 50): "insert",
        (27, 91, 50, 48, 126): "F9",
        (27, 91, 50, 49, 126): "F10",
        (27, 91, 50, 51, 126): "F11",
        # (27, 91, 50, 52, 126): "Cmd Backspace",
        (27, 91, 50, 52, 126): "F12",
        (27, 91, 51): "delete",
        (27, 91, 51, 126): "Del",
        # (27, 91, 51, 126): "FF12",
        (27, 91, 53): "page up",
        (27, 91, 53, 126): "page up",
        (27, 91, 54): "page down",
        (27, 91, 54, 126): "page down",
        (27, 91, 65): "up",
        (27, 91, 66): "down",
        (27, 91, 67): "right",
        (27, 91, 68): "left",
        (27,): "esc",
        (1,): "Ctrl A",
        (32,): "space",
        (48,): "0",
        (65,): "A",
        (97,): "a",
        (127,): "backspace",
    }
    return add_ascii_keys(result)


def add_ascii_keys(data) -> Dict[Tuple[int, ...], str]:
    """Update the data with ascii keys

    >>> data = add_ascii_keys({})
    >>> assert data[(48,)] == '0'
    >>> assert data[(66,)] == 'B'
    >>> assert data[(99,)] == 'c'
    """
    # See previous function for previous key, value pairs
    for i in range(32):
        data[(i + 1,)] = f"Ctrl {chr( ord('A') + i)}"
    for i in range(32, 127):
        data[(i,)] = f"{chr(i)}"
    return data


def _yielder(getter):
    """Keep yielding from that getter until KeyboardInteruptted"""

    def yield_till_stopped():
        """Yield from the getter until KeyboardInteruptted"""
        while True:
            try:
                yield getter()
            except KeyboardInterrupt:
                raise NoKeys

    return yield_till_stopped


yield_keycodes = _yielder(_get_keycodes)
yield_asciis = _yielder(get_ascii)
yield_ASCIIs = _yielder(get_ASCII)
yield_as_keys = _yielder(get_as_key)


def get_string():
    """A better str(_get_keycodes()) method"""
    keycodes = _get_keycodes()
    initial_code, codes = keycodes[0], keycodes[1:]
    initial_char = chr(initial_code)
    if initial_code == 27:
        initial_char = "\\e"
    elif not ascii.isgraph(initial_char):
        initial_char = "\\x%x" % initial_code
    chars = "".join([chr(c) for c in codes])
    return "".join((initial_char, chars))


user = getpass.getuser()


def ask_user_simplified(
    prompt: str = f"Hello {user}",
    default_key: str = "y",
    simplifier: Callable = lambda x: x.lower(),
):
    if prompt or default_key:
        default_prompt = f"[{default_key}] " if default_key else ""
        sys.stdout.write(f"{prompt} {default_prompt}")
        sys.stdout.flush()
    result = simplifier(get_key())
    if not result or result in ('^i', '^j', '^m'):
        result = default_key
    if not result:
        raise KeyboardInterrupt
    print()
    return result



"""Classes for configuration"""

from dataclasses import dataclass

from pysyte.types import paths
from pysyte.types.dictionaries import NameSpaces
from pysyte.types.paths import FileTypes


class Configuration(NameSpaces):
    """We'll use a new name now"""


class PysyteConfiguration(Configuration):
    def __init__(self, stem):
        self.exts = []
        self.path = self.as_path(stem)
        self.data = self.load(self.path) if self.path else {}
        super().__init__(self.data)

    def as_path(self, stem):
        stem_ = paths.path(stem)
        paths_ = [stem_.extend_by(_) for _ in self.exts]
        for path in paths_:
            for ext in self.exts:
                path = path.extend_by(ext)
                if path.isfile():
                    return path
        return paths.path(None)


class YamlConfiguration(PysyteConfiguration):
    """Read a yaml config file and parse it to attributes"""

    def __init__(self, module):
        self.exts = ["yml", "yaml"]
        super().__init__(paths.path(module))

    def load(self, path):
        return yaml_load(path)

    def extensions(self):
        return ("yml", "yaml")


class ModuleConfiguration(YamlConfiguration):
    def __init__(self, module):
        super().__init__(paths.path(module))


@dataclass
class ConfigPathsData:
    paths: list


class ConfigPaths(ConfigPathsData):
    def __init__(self, paths_):
        super().__init__([_ for _ in paths_ if _])
        self.file_types = FileTypes(
            [
                (YamlConfiguration, "yml"),
                (YamlConfiguration, "yaml"),
            ]
        )

    def configs(self, name):
        for path_ in self.paths:
            stem = path_ / name
            for typed in list(self.file_types.typed(stem)):
                if typed:
                    yield typed

    def files(self, name):
        for path_ in self.paths:
            stem = path_ / name
            for file in list(self.file_types.files(stem)):
                if file:
                    yield file

    def load(self, name):
        result = {}
        for config in self.configs(name):
            result.update(config)
        return Configuration(result)



"""Handle configs from program name"""

from typing import Optional
from typing import List

from pysyte import __file__ as path_to_pysyte
from pysyte.oss import linux
from pysyte.config.types import ConfigPaths
from pysyte.types.dictionaries import NameSpaces
from pysyte.types.paths import DirectPath
from pysyte.types.paths import path


def _common_config_dirs(extras: List[DirectPath]) -> List[DirectPath]:
    """Build a list of common config dirs

    Try each of the following directories
        include them if they exist on this machine
        exclude any duplicates
    And in the following order

    /etc/, $XDG_CONFIG_DIRS, ~/.config, $XDG_CONFIG_HOME
        then any extras specified as args
    """

    def add_dir(value: str):
        path_ = path(value)
        if not path_:
            return
        expanded = path_.expand()
        if expanded in configs:
            return
        if expanded.isdir():
            configs.append(expanded)
        elif expanded.isfile():
            configs.append(expanded.parent)

    configs: List[DirectPath] = []
    add_dir("/etc")
    for path_ in linux.xdg_dirs():
        add_dir(path_)
    add_dir("~/.config")
    add_dir(linux.xdg_home())
    for path_ in extras or []:
        add_dir(path_)
    return configs


def load_configs(name: str, extras: Optional[list] = None) -> NameSpaces:
    """Load all config files with that name from common config dirs"""
    config_paths = ConfigPaths(_common_config_dirs(extras or []))
    return config_paths.load(name)


pysyte = load_configs("pysyte", [path(path_to_pysyte)])



"""Handle datetime for pysyte types"""

import datetime
import time


def taken(diff):
    """Convert a time diff to total number of microseconds"""
    microseconds = diff.seconds * 1_000 + diff.microseconds
    return abs(diff.days * 24 * 60 * 60 * 1_000 + microseconds)


def seconds_taken(diff):
    """Convert a time diff to total number of seconds"""
    return int(diff.total_seconds())


def now():
    """Convenience method to call datetime's now"""
    return datetime.datetime.now()


def microseconds_since_epoch():
    """Number of microseconds since the epoch"""
    return int(time.time() * 1_000)


def seconds_since_epoch():
    """Number of seconds since the epoch"""
    return int(time.time())



from contextlib import contextmanager

from rich.console import Console
from rich.traceback import install

install(show_locals=True)


_console = Console()


@contextmanager
def rich_exceptions(locals_: dict):
    if locals_:
        locals().update(locals_)
    try:
        yield
    except:  # noqa
        _console.print_exception(show_locals=True)



"""Methods to provide names for colours

And numbers for those names"""


def bw():
    """The simplest of all - no colour and full colour"""
    return "black white".split()


def rgb():
    """The basic cone sensitivities of your eyes"""
    return "red green blue".split()


primaries = rgb


def cym():
    """The complements of the primaries"""
    return "cyan yellow magenta".split()


secondaries = complementaries = cym


def basic():
    """The basic colour names: Black, the primaries, their complements and white"""
    return bw() + rgb() + cym()


def _cga_colour_names():
    """The basic CGA colours are in a slightly different order"""
    return "red green yellow blue magenta cyan".split()


def dark_cga():
    return ["black"] + _cga_colour_names() + ["silver"]


def light_cga():
    return ["gray"] + [f"light {name}" for name in _cga_colour_names()] + ["white"]


def cga():
    return dark_cga() + light_cga()



"""Generic methods for splitting strings"""

import re
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from pysyte.types.literals import punctuation
from pysyte.types.literals import nones


def _default_separator() -> str:
    """The default separator used by all methods is a comma

    Monkey-patching this method can change that default
    """
    return punctuation.comma


def join(items: list, separator: Optional[str] = None) -> str:
    """Join the items into a string using the separator

    Converts items to strings if needed

    >>> join([1, 2, 3])
    '1,2,3'
    """
    if not items:
        return ""
    if separator is None:
        separator = _default_separator()
    return separator.join([str(item) for item in items])


def seamless_join(items: list) -> str:
    """A seamless join does not show where the joins are

    >>> seamless_join(['fred', 'was', 'here'])
    'fredwashere'
    """
    return join(items, nones.string)


def split(string: str, separator_regexp: Optional[str] = None, maxsplit=0) -> List[str]:
    """Split a string to a list

    >>> split('fred, was, here')
    ['fred', ' was', ' here']
    """
    if not string:
        return []
    if separator_regexp is None:
        separator_regexp = _default_separator()
    if not separator_regexp:
        return string.split()
    return re.split(separator_regexp, string, maxsplit)


def split_and_strip(
    string: str, separator_regexp: Optional[str] = None, maxsplit=0
) -> List[str]:
    """Split a string into items and trim any excess spaces from the items

    >>> split_and_strip('fred, was, here  ')
    ['fred', 'was', 'here']
    """
    if not string:
        return [""]
    if separator_regexp is None:
        separator_regexp = _default_separator()
    if not separator_regexp:
        return string.split()
    return [item.strip() for item in re.split(separator_regexp, string, maxsplit)]


def split_and_strip_without(
    string: str, exclude, separator_regexp: Optional[str] = None
) -> List[str]:
    """Split a string into items, and trim any excess spaces

    Any items in exclude are not in the returned list

    >>> split_and_strip_without('fred, was, here  ', ['was'])
    ['fred', 'here']
    """
    result = split_and_strip(string, separator_regexp)
    if not exclude:
        return result
    return [x for x in result if x not in exclude]


def split_and_strip_whole(
    string: str, separator_regexp: Optional[str] = None
) -> List[str]:
    """Split a string into items and trim any excess spaces from the items

    Exclude any empty items

    >>> split_and_strip_whole('fred, , was,here,')
    ['fred', 'was', 'here']
    """
    return split_and_strip_without(string, [""], separator_regexp)


def split_by_count(items: list, count, filler: Optional[Any] = None) -> List[Tuple]:
    """Split the items into tuples of count items each

    >>> split_by_count([0, 1, 2, 3], 2)
    [(0, 1), (2, 3)]

    If there are a mutiple of count items then filler makes no difference
    >>> split_by_count([0, 1, 2, 7, 8, 9], 3, 0) == split_by_count(
    ...     [0, 1, 2, 7, 8, 9], 3
    ... )
    True

    If there are not a multiple of count items, then any extras are discarded
    >>> split_by_count([0, 1, 2, 7, 8, 9, 6], 3)
    [(0, 1, 2), (7, 8, 9)]

    Specifying a filler expands the "lost" group
    >>> split_by_count([0, 1, 2, 7, 8, 9, 6], 3, 0)
    [(0, 1, 2), (7, 8, 9), (6, 0, 0)]
    """
    if filler is not None:
        items = items[:]
        while len(items) % count:
            items.append(filler)
    iterator = iter(items)
    iterators = [iterator] * count
    return list(zip(*iterators))


def pairs(items: list, filler: Optional[Any] = None) -> List[Tuple]:
    """Split the items into pairs

    >>> pairs([0, 1, 2, 7, 8, 9, 10])
    [(0, 1), (2, 7), (8, 9)]

    If a filler is used any unpaired items are retained
    >>> pairs([0, 1, 2, 7, 8, 9, 10], 99)
    [(0, 1), (2, 7), (8, 9), (10, 99)]
    """
    return split_by_count(items, 2, filler)


def threes(items: list, filler: Optional[Any] = None) -> List[Tuple]:
    """Split the items into groups of 3

    >>> threes([0, 1, 2, 6, 7, 8, 9])
    [(0, 1, 2), (6, 7, 8)]

    If a filler is used any discarded items are retained
    >>> threes([0, 1, 2, 6, 7, 8, 9], 5)
    [(0, 1, 2), (6, 7, 8), (9, 5, 5)]
    """
    return split_by_count(items, 3, filler)


def despaced(string: str) -> List[str]:
    """Split a string into spaceless items

    Split on spaces, trim excess space, exclude any empty strings

    >>> despaced('fred, , was,here today')
    ['fred,', ',', 'was,here', 'today']
    """
    return split_and_strip_without(string, [""], " ")


def words(string: str) -> List[str]:
    """Split a string into words

    Split on (English) punctuaution, trim space, exclude any empty strings

    >>> words('fred, , was,here today')
    ['fred', 'was', 'here', 'today']
    """
    return split_and_strip_without(string, [""], "[,;. ]")


def rejoin(string: str, separator_regexp: Optional[str] = None, spaced=False) -> str:
    """Split a string and then rejoin it

    Spaces are interspersed between items only if spaced is True

    >>> rejoin('fred, was, here  ')
    'fred,was,here'
    """
    strings = split_and_strip(string)
    if separator_regexp is None:
        separator_regexp = _default_separator()
    joiner = spaced and f"{separator_regexp} " or separator_regexp
    return joiner.join(strings)



from pysyte import os

def exits():
    import os
    return {k:v for k, v in os.globals().items() if k.startswith("EX_")}

@dataclass
class ExitCode:
    exit_code: int = EX_OK

    def __post_init__(self):
        self.exit = self.string()

    def __int__(self) -> int:
        return self.exit_code

    def __str__(self):
        return self.exit

    @property
    def ok(self):
        return self.exit_code == EX_OK

    @property
    def errors(self):
        return not self.ok

    def string(self) -> str:
        if self.ok:
            return "EX_OK"
        import os
        for ex_name, code in exits().items():
            if code == self.exit_code:
                return ex_name
        exit = self.exit_code
        return f"{exit=}"

pass_ = os.ExitCode(os.EX_OK)
fail = os.ExitCode(os.EX_FAIL)




from pysyte.cli.arguments import ArgumentsParser
from pysyte.types.paths import path


class PathParser(ArgumentsParser):
    """Handle positional args as paths"""

    def __init__(self, argparser, paths_arg=None):
        super().__init__(argparser)
        self.paths_arg = paths_arg if paths_arg else "paths"
        super().positionals(self.paths_arg)

    def post_parser(self, arguments):
        strings = arguments.get_arg(self.paths_arg)
        paths = [path(_) for _ in strings]
        arguments.set_arg(self.paths_arg, [_ for _ in paths if _])
        arguments.set_arg(f"not_{self.paths_arg}", [_ for _ in paths if not _])
        return arguments


def add_args(old_parser, name=None):
    parser = PathParser(old_parser.parser, name)
    return parser



"""Handle some random choices"""
import random

from pysyte.cli import app


def flip() -> bool:
    return random.choice((True, False))


def coin() -> str:
    return "heads" if flip() else "tails"


def main():
    result = coin()
    assert result == "heads", f"{result=}"


if __name__ == "__main__":
    app.exit(main)



"""The Unix filesystem

https://en.wikipedia.org/wiki/Unix_filesystem#Conventional_directory_layout

The filesystem appears as one rooted tree of directories

Instead of addressing separate volumes such as disk partitions,
    removable media,
    and network shares as separate trees (
        as done in DOS and Windows:
            each drive has a LETTER naming the root of its file system tree),
    such volumes can be mounted on a directory,
    causing the volume's tree to appear as that directory in the larger tree
    The root of the entire tree is denoted.
"""

from pysyte.config.types import ModuleConfiguration
from pysyte.types.paths import DirectPath


class UnixDirectory(DirectPath):
    def __init__(self, path_):
        super().__init__(path_)


def upath(string):
    """For consistency with path()"""
    return UnixDirectory(string)


root = upath("/")
config = ModuleConfiguration(__file__)



"""Some popular similarity measures in python

https://dataaspirant.com/2015/04/11/five-most-popular-similarity-measures-implementation-in-python/

Similarity:

The similarity measure is the measure of how much alike two data objects are.
    Similarity measure in a data mining context is a distance with dimensions
    representing features of the objects. If this distance is small, it will be
    the high degree of similarity where large distance will be the low degree
    of similarity.

The similarity is subjective and is highly dependent on domain and application.
    For example, two fruits are similar because of color or size or taste.
    Care should be taken when calculating distance across dimensions/features
    that are unrelated. The relative values of each element must be normalized,
    or one feature could end up dominating the distance calculation. Similarity
    are measured in the range 0 to 1 [0,1].

"""

from math import sqrt
from decimal import Decimal


def euclidean(x: list, y: list) -> float:
    """Euclidean distance is the most common use of distance.

    Euclidean distance is also known as simply distance.
    When data is dense or continuous, this is the best proximity measure.

    The Euclidean distance between two points is the length connecting them.
    The Pythagorean theorem gives this distance between two points.
    """
    return sqrt(sum(pow(a - b, 2) for a, b in zip(x, y)))


def manhattan(x: list, y: list) -> float:
    """Manhattan is the the sum of the absolute diff of their Cartesian co-ords

    In a plane with p1 at (x1, y1) and p2 at (x2, y2).
        Manhattan distance = |x1 – x2| + |y1 – y2|

    This Manhattan distance metric is also known as
        Manhattan length,
        rectilinear distance,
        L1 distance or L1 norm,
        city block distance,
        Minkowski's L1 distance,
        taxi-cab metric
    """
    return sum(abs(a - b) for a, b in zip(x, y))


def minkowski(x: list, y: list, p_value: float) -> float:
    """Minkowski distance is general form of Euclidean and Manhattan distances

    In the equation,
        d^MKD is the Minkowski distance between the data record i and j,
        k the index of a variable, n the total number of variables y and
        λ the order of the Minkowski metric.

    Although it is defined for any λ > 0, it is rarely used except 1, 2 and ∞.

    The way distances are measured by the Minkowski metric of different orders
        between two objects with three variables ( In the image it displayed
        in a coordinate system with x, y ,z-axes).

    Synonyms of Minkowski:

    λ = 1 is the Manhattan distance.
    λ = 2 is the Euclidean distance.
        Synonyms are L2-Norm or Ruler distance.
        For two vectors of ranked ordinal variables
            the Euclidean distance is sometimes called Spear-man distance.
    λ = ∞ is the Chebyshev distance.
        Synonyms are Lmax-Norm or Chessboard distance.
    """

    def nth_root(value: float, n_root: float) -> float:
        root_value = 1 / float(n_root)
        return float(round(Decimal(value) ** Decimal(root_value), 3))

    return nth_root(sum(pow(abs(a - b), p_value) for a, b in zip(x, y)), p_value)



"""Methods to handle streams"""


import sys
from contextlib import contextmanager
from typing import Optional
from typing import TextIO
from typing import Tuple
from typing import Generator

from six import StringIO


@contextmanager
def swallow_stdout(stream: Optional[TextIO] = None) -> Generator[TextIO, None, None]:
    """Divert stdout into the given stream

    >>> with swallow_stdout() as stream:
    ...     print('hello', end='')
    ...
    >>> assert stream.getvalue() == 'hello'
    """
    saved = sys.stdout
    if stream is None:
        stream = StringIO()
    sys.stdout = stream
    try:
        yield stream
    finally:
        sys.stdout = saved


@contextmanager
def swallow_stderr(stream: Optional[TextIO] = None) -> Generator[TextIO, None, None]:
    """Divert stdout into the given stream

    >>> with swallow_stderr() as string:
    ...     print('hello', end='', file=sys.stderr)
    ...
    >>> assert string.getvalue() == 'hello'
    """
    saved = sys.stderr
    if stream is None:
        stream = StringIO()
    sys.stderr = stream
    try:
        yield stream
    finally:
        sys.stderr = saved


@contextmanager
def swallow_std() -> Generator[Tuple[TextIO, TextIO], None, None]:
    """Divert stdout and stderr to the given stream

    >>> with swallow_std() as streams:
    ...     print('hello', end=' ', file=sys.stderr)
    ...     print('world', end='', file=sys.stderr)
    ...
    >>> assert streams[0].getvalue() + streams[1].getvalue() == 'hello world'
    """
    saved = sys.stdout, sys.stderr
    out, err = StringIO(), StringIO()
    sys.stdout = out
    sys.stderr = err
    try:
        yield out, err
    finally:
        sys.stdout, sys.stderr = saved



"""A collection of extended path classes and methods

The classes all inherit from the original path.path
"""
from __future__ import annotations

import os
import re
import stat
import sys
from dataclasses import dataclass
from fnmatch import fnmatch
from functools import singledispatch
from importlib import import_module
from typing import Iterable
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

from deprecated import deprecated

from path import Path as path_Path
from pysyte.types.lists import flatten
from pysyte.types.methods import Method


class PathError(Exception):
    """Something went wrong with a path"""

    prefix = "Path Error"


class MissingPath(PathError):
    def __init__(self, path, desc=""):
        self.path = path
        description = desc or "path"
        super().__init__(f"Missing {description}{path}")


class MissingImport(MissingPath):
    def __init__(self, module):
        self.module = module
        try:
            path_ = module.__file__
        except AttributeError:
            path_ = module.__name__
        super().__init__(path_, desc="module")


class PathAssertions:
    """Assertions that can be made about paths"""

    def assertExists(self):
        if not self.exists():
            raise MissingPath(self)
        return self

    def assert_isdir(self):
        """Raise a PathError if this path is not a directory on disk"""
        if not self.isdir():
            raise PathError(f"{self} is not a directory")
        return self

    def assert_isfile(self):
        """Raise a PathError if this path is not a file on disk"""
        if not self.isfile():
            raise PathError(f"{self} is not a file")
        return self


StrPath = Union["StringPath", str]  # many args can be either string or path


class StringPath(path_Path):
    """This class handles the path as if it were just a string

    Sub-classes know about paths qua paths
    """

    # pylint: disable=abstract-method
    # pylint: disable=too-many-public-methods

    def __hash__(self):
        return hash(str(self))

    def __repr__(self) -> str:
        string = repr(f"{self}")
        return f"<{self.__class__.__name__} {string}>"

    @deprecated(version="0.7.5", reason="Please use Python 3")
    def __div__(self, other) -> StringPath:
        """Need to cover parent's use of older idiom"""
        return self.__truediv__(other)

    def __truediv__(self, substring: str) -> StringPath:
        """Handle the / operator

        Add substring to self

        >>> p = StringPath("/path/to")
        >>> assert p.__truediv__("fred") == p / "fred"
        >>> assert p / "fred" == "/path/to/fred"
        >>> assert p / None is p
        """
        if not substring:
            return self
        full_string = os.path.join(str(self), substring)
        return makepath(full_string)

    def __floordiv__(self, substrings: Sequence[str]) -> StringPath:
        """Handle the // operator

        Add substrings to self like a path in local os

        >>> p = StringPath("/path/to")
        >>> assert p.__floordiv__("fred") == p // "fred"
        >>> assert p // ["module", "fred.py"] == "/path/to/module/fred.py"
        >>> assert p // None is p
        """
        if not substrings:
            return self
        string = str(self)
        strings = [string] + list(substrings)
        return makepath(os.path.join(*strings))

    def __eq__(self, other) -> bool:
        return str(self) == str(other)

    def __lt__(self, other) -> bool:
        return str(self) < str(other)

    # functools.total_ordering does not work properly cos we inherit from str
    # Hence: we do need to define next 3
    def __le__(self, other):
        return self.__eq__(other) or self.__lt__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __contains__(self, other) -> bool:
        """The other is in self if self.contains(other)

        this method should be specialised in sub-classes
        """
        return self.contains(other)

    def contains(self, other: StrPath) -> bool:
        """If other is also a path then this path should start with other

        E.g. /path/to/file is "in" /path

        Otherwise, just use the sub-string sense of "in"
        """
        if isinstance(other, StringPath):
            return str(other).startswith(str(self))
        return str(other) in str(self)

    def basename(self) -> str:
        return str(super().basename())

    @property
    def basename_(self) -> str:
        return self.basename()

    @property
    def name(self) -> str:
        return str(super().name)

    @property
    def stem(self) -> StringPath:
        stem, *_ = self.splitexts()
        return stem

    @property
    def stem_name(self) -> str:
        return self.stem.name

    def splitexts(self) -> Tuple[StringPath, str]:
        """Split all extensions from the path

        >>> p = FilePath('here/fred.tar.gz')
        >>> assert p.splitexts() == ('here/fred', '.tar.gz')
        """
        copy = self[:]
        filename, ext = os.path.splitext(copy)
        zippers = (
            ".gz",
            ".bz",
            ".zip",
            ".bzip",
        )
        for zipper in zippers:
            if ext == zipper:
                filename, ext_ = os.path.splitext(filename)
                ext = f"{ext_}{zipper}"
        return self.__class__(filename), ext

    def add_ext(self, *args) -> StringPath:
        """Join all args as extensions

        Strip any leading `.` from args

        >>> source = makepath(__file__)
        >>> new = source.add_ext('txt', '.new')
        >>> assert new.name.endswith('.py.txt.new')
        """
        exts = [(a[1:] if a[0] == "." else a) for a in args]
        string = ".".join([self] + list(exts))
        return makepath(string)

    def add_missing_ext(self, ext: str) -> StringPath:
        """Add that extension, if it is missing

        >>> fred = makepath("fred")
        >>> assert fred.add_missing_ext("") == fred
        >>> fred_py = makepath("fred.py")
        >>> assert fred.add_missing_ext(".py") == fred_py
        >>> assert fred_py.add_missing_ext(".txt") == "fred.py.txt"
        """
        dot_ext = f'.{ext.lstrip(".")}'
        copy = self[:]
        _, self_ext = os.path.splitext(copy)
        return makepath(self) if self_ext == dot_ext else self.add_ext(dot_ext)

    def extend_by(self, ext: str) -> StringPath:
        """The path to the file changed to use the given ext

        >>> fred = "/path/to/fred.fred"
        >>> assert makepath("/path/to/fred").extend_by("fred") == fred
        >>> assert makepath("/path/to/fred.txt").extend_by(".fred") == fred
        >>> assert makepath("/path/to/fred.txt").extend_by("..fred") == fred
        """
        copy = self[:]
        filename, _ = os.path.splitext(copy)
        ext_ = ext.lstrip(".")
        return makepath(f'{filename}.{ext_}')

    def has_vcs_dir(self):
        for vcs_dir in (".git", ".svn", ".hg"):
            if self.fnmatch_part(vcs_dir):
                return True
        return False


def ext_language(ext, exts=None, simple=True):
    """Language of the extension in those extensions

    If exts is supplied, then restrict recognition to those exts only
    If exts is not supplied, then use all known extensions

    >>> ext_language('.py') == 'python'
    True
    """
    languages = {
        ".py": "python",
        ".py2": "python" if simple else "python2",
        ".py3": "python" if simple else "python3",
        ".sh": "bash",
        ".bash": "bash",
        ".pl": "perl",
        ".txt": "english",
    }
    ext_languages = {_: languages[_] for _ in exts} if exts else languages
    return ext_languages.get(ext)


class NonePath(StringPath):
    def __init__(self, string=None):
        self.string = string if string else ""
        self.proxy = FilePath(self.string) or DirectPath(self.string)

    def __str__(self):
        return self.string

    def __repr__(self):
        string = self.string
        if self.proxy:
            string = str(self.proxy)
        return f'<{self.__class__.__name__} "{string}">'

    def __bool__(self):
        return False

    def __eq__(self, other):
        if self.string:
            return str(self) == str(other)
        return not other

    def __lt__(self, other):
        if self.string and other:
            return str(self) < str(other)
        return bool(other)

    def contains(self, other: StrPath) -> bool:
        """As this is not a real path, just use the substring sense"""
        return str(other) in str(self)

    def __truediv__(self, child):
        result = os.path.join(self.string, child) if child else self.string
        return makepath(result)

    @property
    def parent(self):
        if "/" not in self.string:
            return None
        parent_string = "/".join(self.string.split("/")[:-1])
        return makepath(parent_string)

    def exists(self):
        return False

    isdir = isfile = isexec = isroot = exists

    def __getattr__(self, name):
        return getattr(self.proxy, name, None)

    def makedirs(self):
        os.makedirs(str(self))


class DotPath(StringPath):
    """This class add path-handling to a string"""

    def parent_directory(self):
        if self.isroot():
            return None
        return self.parent

    def parent_directories(self):
        if self.isroot():
            return []
        parent = self.parent
        return [parent] + parent.parent_directories()

    def directory(self):
        """Return a path to the path's directory"""
        return self.parent

    def dirnames(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> DotPath('/path/to/module.py').dirnames() == ['/', 'path', 'to']
        True
        >>> DotPath('path/to/module.py').dirnames() == ['path', 'to']
        True
        """
        return [str(_) for _ in self.directory().split(os.path.sep)]

    def dirpaths(self):
        """Split the dirname into individual directory names

        An absolute path starts with an empty string, a relative path does not

        >>> p = DotPath('/path/to/x.py')
        >>> assert p.paths == p.dirpaths()
        """
        parts = self.split()
        result = [DotPath(parts[0] or "/")]
        for name in parts[1:]:
            result.append(result[-1] / name)
        return result

    def directories(self):
        """Split the dirname into individual directory names

        No empty parts are included

        >>> DotPath('path/to/module.py').directories() == ['path', 'to']
        True
        >>> DotPath('/path/to/module.py').directories() == ['/', 'path', 'to']
        True
        """
        return [d for d in self.dirnames() if d]

    parents = property(
        dirnames,
        None,
        None,
        """ This path's parent directories, as a list of strings.

        >>> DotPath('/path/to/module.py').parents == ['/', 'path', 'to']
        True
        """,
    )

    paths = property(
        dirpaths,
        None,
        None,
        """ This path's parent directories, as a sequence of paths.

        >>> paths = DotPath('/usr/bin/vim').paths
        >>> assert paths[-1].exists()  # vim might be a link
        >>> assert paths[-2] == paths[-1].parent
        >>> assert paths[-3] == paths[-2].parent
        >>> assert paths[-4] == paths[-3].parent
        >>> assert paths[-4] == paths[0]
        """,
    )

    def path_split(self, sep=None, maxsplit=-1):
        separator = sep or os.path.sep
        parts = super().split(separator, maxsplit)
        parts[0] = makepath(parts[0] if parts[0] else "/")
        return parts

    split = path_split

    def abspath(self):
        return makepath(os.path.abspath(str(self)))

    def slashpath(self):
        return self + "/" if self.isdir() else self

    def short_relative_path_to(self, destination):
        """The shorter of either the absolute path of the destination,
            or the relative path to it

        >>> print(
        ...     DotPath('/home/guido/bin').short_relative_path_to(
        ...         '/home/guido/build/python.tar'
        ...     )
        ... )
        ../build/python.tar
        >>> print(
        ...     DotPath('/home/guido/bin').short_relative_path_to(
        ...         '/mnt/guido/build/python.tar'
        ...     )
        ... )
        /mnt/guido/build/python.tar
        """
        relative = self.relpathto(destination)
        absolute = self.__class__(destination).abspath()
        if len(str(relative)) < len(str(absolute)):
            return relative
        return absolute

    def short_relative_path_to_here(self):
        """A short path relative to current working directory"""
        return self.short_relative_path_to(os.getcwd())

    def short_relative_path_from_here(self):
        """A short path relative to self to the current working directory"""
        return self.__class__(os.getcwd()).short_relative_path_to(self)

    def fnmatch_basename(self, glob):
        if glob.startswith(os.path.sep):
            glob = glob.lstrip(os.path.sep)
        string = self.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directory(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        if self.isdir():
            string = self.basename()
        else:
            string = self.parent.basename()
        if fnmatch(string, glob):
            return self
        return None

    def fnmatch_directories(self, glob):
        if glob.startswith(os.path.sep) or glob.endswith(os.path.sep):
            glob = glob.strip(os.path.sep)
        strings = reversed(self.directory().splitall()[1:])
        for string in strings:
            if fnmatch(string, glob):
                return self
        return None

    def fnmatch_part(self, glob):
        if self.fnmatch(glob):
            return self
        if self.fnmatch_basename(glob):
            return self
        if self.fnmatch_directory(glob):
            return self
        if self.fnmatch_directories(glob):
            return self
        return None

    def expand(self):
        """Expand the path completely

        This removes any "~" (for home dir) and any shell variables
            eliminates any symbolic links
            and converts from relative to absolute path
        """
        u = os.path.expanduser(str(self))
        v = os.path.expandvars(u)
        r = os.path.realpath(v)
        return makepath(r)

    def same_path(self, other):
        """Whether this path points to same place as the other"""
        return self.expand() == other.expand()

    def isroot(self):
        raise NotImplementedError("Only a directory path can be a root")

    def ishidden(self):
        """A 'hidden file' has a name starting with a '.'"""
        s = str(self.basename())
        return s and s[0] == "."

    def is_executable(self):
        """Whether the path is executable"""
        # pylint: disable=no-self-use
        return False

    def isexec(self):
        return self.is_executable()

    def has_executable(self):
        """Whether the path has any executable bits set"""
        executable_bits = stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH
        try:
            return bool(os.stat(self).st_mode & executable_bits)
        except OSError:
            return False


class FilePath(DotPath, PathAssertions):
    """A path to a known file"""

    def __truediv__(self, child):
        raise PathError("%r has no children" % self)

    def __iter__(self):
        for line in self.stripped_lines():
            yield line

    def contains(self, other: StrPath) -> bool:
        """Whether other is in this file's text"""
        return str(other) in self.text()

    def stripped_lines(self):
        """A list of all lines without trailing whitespace

        If lines can not be read (e.g. no such file) then an empty list
        """
        try:
            return [_.rstrip() for _ in self.lines(retain=False)]
        except (OSError, IOError, UnicodeDecodeError):
            return []

    def stripped_whole_lines(self):
        """A list of all lines without trailing whitespace or blank lines"""
        return [_ for _ in self.stripped_lines() if _]

    def non_comment_lines(self):
        """A list of all non-empty, non-comment lines"""
        return [_ for _ in self.stripped_whole_lines() if not _.startswith("#")]

    def isroot(self):
        """A file cannot be root of a filesystem"""
        return False

    def is_executable(self):
        """Whether the file has any executable bits set"""
        return self.has_executable()

    def has_line(self, string):
        for line in self:
            if string == line:
                return True
        return False

    def any_line_has(self, string):
        for line in self:
            if string in line:
                return True
        return False

    def as_python(self):
        """The path to the file with a .py extension

        >>> assert FilePath("/path/to/fred.txt").as_python() == "/path/to/fred.py"
        """
        return self.extend_by(".py")

    def make_read_only(self):
        """chmod the file permissions to -r--r--r--"""
        self.chmod(ChmodValues.readonly_file)

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(self.parent)

    def dirname(self):
        return DirectPath(os.path.dirname(self))

    parent = property(dirname)

    def shebang(self):
        """The  #! entry from the first line of the file

        If no shebang is present, return an empty string
        """
        try:
            first_line = self.stripped_lines()[0]
            if first_line.startswith("#!"):
                return first_line[2:].strip()
        except IndexError:
            pass
        return ""

    def mv(self, destination):  # pylint: disable=invalid-name
        return self.move(destination)

    @property
    def language(self):
        """The language of this file"""
        try:
            return self._language
        except AttributeError:
            self._language = ext_language(self.ext)
        return self._language

    @language.setter
    def language(self, value):
        self._language = value


class DirectPath(DotPath, PathAssertions):
    """A path which knows it might be a directory

    And that files are in directories
    """

    __file_class__ = FilePath

    def __iter__(self):
        for a_path in self.listdir():
            yield a_path

    def contains(self, other: StrPath) -> bool:
        """If other is a path then use that sense of "in"

        So /path/to/here is "in" /path

        Otherwise see if other is listed "in" this directory
        """
        if isinstance(other, DotPath):
            return self in other.parent_directories()
        return str(other) in [_.name for _ in self.listdir()] + [".", ".."]

    def directory(self):
        """Return a path to a directory.

        Either the path itself (if it is a directory), or its parent)
        """
        if self.isdir():
            return self
        return self.parent

    def remove_dir(self):
        """Try to remove the path

        If it is a directory, try recursive removal of contents too
        """
        if self.islink():
            self.unlink()
        elif self.isdir():
            self.empty_directory()
            if self.isdir():
                self.rmdir()
        else:
            return False
        return True

    def empty_directory(self):
        """Remove all contents of a directory

        Including any sub-directories and their contents"""
        for child in self.walkfiles():
            child.remove()
        for child in reversed([self.walkdirs()]):
            if child == self or not child.isdir():
                continue
            child.rmdir()

    def cd(self):  # pylint: disable=invalid-name
        """Change program's current directory to self"""
        return cd(StringPath(self))

    def list_dirs(self, pattern=None):
        return self.list_dirs_files(pattern)[0]

    def list_files(self, pattern=None):
        return self.list_dirs_files(pattern)[1]

    def list_dirsfiles(self, pattern=None):
        dirs, others = self.list_dirs_files(pattern)
        return dirs + others

    def list_dirs_files(self, pattern=None):
        items = self.listdir(pattern)
        dirs = [_ for _ in items if _.isdir()]
        others = [_ for _ in items if not _.isdir()]
        return dirs, others

    def make_read_only(self):
        """chmod the directory permissions to -r-xr-xr-x"""
        self.chmod(ChmodValues.readonly_directory)

    def touch_file(self, filename):
        """Touch a file in the directory"""
        path_to_file = self.__file_class__(os.path.join(self, filename))
        path_to_file.touch()
        return path_to_file

    def existing_sub_paths(self, sub_paths):
        """Those in the given list of sub_paths which do exist"""
        paths_to_subs = [self / _ for _ in sub_paths]
        return [_ for _ in paths_to_subs if _.exists()]

    # pylint: disable=arguments-differ
    def walkdirs(self, pattern=None, errors="strict", ignores=None):
        ignored = ignore_fnmatches(ignores)
        for path_to_dir in super(DirectPath, self).walkdirs(pattern, errors):
            if not ignored(path_to_dir.relpath(self)):
                yield path_to_dir

    # pylint: disable=arguments-differ
    def walkfiles(self, pattern=None, errors="strict", ignores=None):
        ignored = ignore_fnmatches(ignores)
        for path_to_file in super(DirectPath, self).walkfiles(pattern, errors):
            if not ignored(path_to_file.relpath(self)):
                yield path_to_file

    def listfiles(self, pattern=None, ignores=None):
        ignored = ignore_fnmatches(ignores)
        return [_ for _ in self.listdir(pattern) if _.isfile() and not ignored(_)]

    def isroot(self):
        return str(self) == "/"


@dataclass
class FileTypeData:
    type_: type
    ext: str


class FileType(FileTypeData):
    def file(self, path_):
        return path_.add_missing_ext(self.ext)

    def typed(self, path_):
        file = self.file(path_)
        if file:
            return self.type_(file)
        return None


@dataclass
class FileTypesData:
    file_types: List[FileType]


class FileTypes(FileTypesData):
    def types(self):
        types_ = self.file_types
        return [_ if isinstance(_, FileType) else FileType(*_) for _ in types_]

    def files(self, path_):
        for file_type in self.types():
            file = file_type.file(path_)
            if file:
                yield file

    def typed(self, path_):
        for file_type in self.types():
            typed = file_type.typed(path_)
            if typed:
                yield typed


def ignore_fnmatches(ignores):
    def ignored(a_path):
        if not ignores:
            return False
        for ignore in ignores:
            if a_path.fnmatch_part(ignore):
                return True
        return False

    return ignored


class ChmodValues:
    # pylint: disable=too-few-public-methods
    readonly_file = 0o444
    readonly_directory = 0o555


def _make_module_path(arg):
    """Make a path from a thing that has a module

    classes and functions have modules, they'll be needing this
    """
    try:
        return makepath(import_module(arg.__module__))
    except (AttributeError, ModuleNotFoundError):
        return None


@singledispatch
def makepaths(arg) -> Paths:
    attribute = getattr(arg, "paths", [])
    return makepaths(attribute) if attribute else makepaths(list(arg))


@singledispatch
def makepath(arg) -> StringPath:
    attribute = getattr(arg, "path", "")
    return makepath(attribute) if attribute else makepath(str(arg))


path = makepath


@makepath.register(type(None))
def _mp(arg) -> StringPath:
    """In the face of ambiguity, refuse the temptation to guess."""
    return NonePath()


@makepath.register(DotPath)
def __mp(arg) -> StringPath:
    return arg


@makepath.register(str)
def ____mp(arg) -> StringPath:
    """Make a path from a string

    Expand out any variables, home squiggles, and normalise it
    See also http://stackoverflow.com/questions/26403972

    See also Lynton Kwesi Johnson:
        The Eagle and The Bear have people living in fear
        Of impending nuclear warfare
    """
    if not arg:
        return makepath(None)
    if os.path.isfile(arg):
        return FilePath(arg)
    if os.path.isdir(arg):
        string = arg if arg == "/" else arg.rstrip("/")
        return DirectPath(string)
    v = os.path.expandvars(arg)
    u = os.path.expanduser(v)
    if arg == u:
        return NonePath(arg)
    if os.path.exists(u):
        return makepath(u)
    return NonePath(arg)


def imports():
    return {sys, os, re, stat}


@makepath.register(type(os))
def _____mp(arg) -> StringPath:
    """Make a path from a module"""
    if arg.__name__ == "builtins":
        return NonePath("builtins")
    try:
        return makepath(arg.__file__)
    except AttributeError:
        if arg not in imports() and arg not in sys.path:
            raise MissingImport(arg)
        python_ = makepath(sys.executable)
        assert str(python_.parent.name) == "bin"
        bin_ = python_.parent
        root_ = bin_.parent
        lib = root_ / "lib"
        assert lib.isdir()
        module = lib / f"{arg}.py"
        if module.isfile():
            return module
        package = lib / arg
        if package.isdir():
            return package
        raise ModuleNotFoundError(arg)


@makepath.register(type(makepath))
def ______mp(arg) -> StringPath:
    """Make a path from a function's module"""
    method = Method(arg)
    stdin_regexp = re.compile("<(stdin|.*python-input.*)>")
    if stdin_regexp.match(method.filename):
        return NonePath(method.filename)
    return _make_module_path(method)


@makepath.register(type(DotPath))  # type: ignore[no-redef]
def _______mp(arg) -> StringPath:
    """Make a path from a class's module"""
    return _make_module_path(arg)


@deprecated(reason="use pathstr()", version="0.7.57")
def makestr(string: str) -> StringPath:
    return pathstr(string)


def pathstr(string: str) -> StringPath:
    """Make a path from a string"""
    if os.path.isfile(string) or os.path.isdir(string):
        return makepath(string)
    return NonePath(string)


def cd(path_to: StringPath) -> bool:
    """cd to the given path

    If the path is a file, then cd to its parent directory

    Remember current directory before the cd
        so that we can cd back there with cd('-')
    """
    if path_to == "-":
        previous = getattr(cd, "previous", "")
        if not previous:
            raise PathError("No previous directory to return to")
        return cd(makepath(previous))
    if not hasattr(path_to, "cd"):
        path_to = makepath(path_to)
    try:
        previous = os.getcwd()
    except OSError as e:
        if "No such file or directory" not in str(e):
            raise
        previous = ""
    if path_to.isdir():
        cd_path = path_to
    elif path_to.isfile():
        cd_path = path_to.parent
    elif not path_to.exists():
        return False
    else:
        raise PathError(f"Cannot cd to {path_to}")
    cd.previous = previous  # type: ignore
    os.chdir(cd_path)
    return True


try:
    setattr(cd, "previous", os.getcwd())
except (OSError, AttributeError):
    setattr(cd, "previous", "")


def as_path(string_or_path):
    """Return the argument as a DirectPath

    If it is already one, return it unchanged
    If not, return the makepath()
    """
    if isinstance(string_or_path, DirectPath):
        return string_or_path
    return makepath(string_or_path)


def string_to_paths(string) -> List[StringPath]:
    for c in ":, ;":
        if c in string:
            return strings_to_paths(string.split(c))
    return [makepath(string)]


def strings_to_paths(strings) -> List[StringPath]:
    return [makepath(s) for s in strings]


def choose_paths(*strings, chooser) -> List[StringPath]:
    return [_ for _ in strings_to_paths(strings) if chooser(_)]


def paths(*strings: Iterable) -> List[StringPath]:
    return choose_paths(*strings, chooser=lambda p: p.exists())


def directories(*strings: Iterable) -> List[StringPath]:
    return choose_paths(*strings, chooser=lambda p: p.isdir())


def files(*strings: Iterable) -> List[StringPath]:
    return choose_paths(*strings, chooser=lambda p: p.isfile())


def root():
    return makepath("/")


def tmp():
    return makepath("/tmp")


def home():
    _home = makepath(os.path.expanduser("~"))
    assert _home
    _ = _home.expand()
    return _home


def pwd():
    return makepath(os.getcwd())


def first_dir(path_string):
    """Get the first directory in that path

    >>> first_dir('usr/local/bin') == 'usr'
    True
    """
    parts = path_string.split(os.path.sep)
    return parts[0]


def first_dirs(path_strings):
    """Get the roots of those paths

    >>> first_dirs(['usr/bin', 'bin']) == ['usr', 'bin']
    True
    """
    return [first_dir(_) for _ in path_strings]


def unique_first_dirs(path_strings):
    """Get the unique roots of those paths

    >>> unique_first_dirs(['usr/local/bin', 'bin']) == set(['usr', 'bin'])
    True
    """
    return set(first_dirs(path_strings))


def paths_in_directory(path_: StringPath) -> List[DotPath]:
    """Get all items in the given directory

    Swallow errors to give an empty list
    """
    try:
        return [_ for _ in path_.listdir() if _]
    except OSError:
        return []


def list_matcher(pattern, path_to_directory, wanted):
    """Gives a method to check if an item matches the pattern, and is wanted

    If path_to_directory is given then items are checked there
    By default, every item is wanted
    """
    return lambda x: fnmatch(x, pattern) and wanted(os.path.join(path_to_directory, x))


def list_items(path_to_directory, glob, wanted=lambda x: True) -> List[StringPath]:
    """All items in the given path which match the given glob and are wanted

    By default, every path is wanted
    """
    path_to_directory_ = makepath(path_to_directory)
    if not path_to_directory_:
        return []
    result = []
    for path_to_item in path_to_directory_.listdir():
        if not fnmatch(path_to_item.name, glob):
            continue
        if wanted(path_to_item):
            result.append(path_to_item)
    return result


def list_sub_directories(path_to_directory, glob):
    """All sub-directories of the given directory matching the given glob"""
    return list_items(path_to_directory, glob, os.path.isdir)


def set_items(path_to_directory, glob, wanted):
    return set(list_items(path_to_directory, glob, wanted))


def set_files(path_to_directory, glob):
    """A list of all files in the given directory matching the given glob"""
    return set_items(path_to_directory, glob, os.path.isfile)


def contains_glob(path_to_directory, glob, wanted=lambda x: True):
    """Whether the given path contains an item matching the given glob"""
    return any(list_items(path_to_directory, glob, wanted))


def contains_directory(path_to_directory, glob):
    """Whether the given path contains a directory matching the given glob"""
    return contains_glob(path_to_directory, glob, os.path.isdir)


def contains_file(path_to_directory, glob):
    """Whether the given directory contains a file matching the given glob"""
    return contains_glob(path_to_directory, glob, os.path.isfile)


def environ_paths(key, default=None):
    default_ = default or ""
    return [makepath(_) for _ in os.environ.get(key, default_).split(":")]


def environ_path(key, default=None):
    default_ = default or ""
    return makepath(os.environ.get(key, default_))


def default_environ_path(key, default):
    return makepath(os.environ.get(key, default))


def add_star(string):
    """Add '*' to string

    >>> assert add_star('fred') == 'fred*'
    """
    return f"{string}*"


def add_stars(strings):
    """Add '.*' to each string

    >>> assert add_stars(['fred', 'Fred.']) == ['fred*', 'Fred.*']
    """
    paths_ = [strings] if isinstance(strings, str) else strings
    return [add_star(p) for p in paths_]


def tab_complete(strings, globber=add_stars, select=os.path.exists):
    """Finish path names "left short" by bash's tab-completion

    strings is a string or strings
        if any string exists as a path return those as paths

    globber is a method which should return a list of globs
        default: add_stars(['fred.']) == ['fred.*']
            or, e.g: add_python(['fred', 'fred.']) == ['fred.py*']
        if any expanded paths exist return those

    select is a method to choose wanted paths
        Defaults to selecting existing paths
    """
    strings_ = [strings] if isinstance(strings, str) else strings
    globs = flatten([globber(s) for s in strings_])
    here_ = pwd()
    matches = []
    for glob_ in globs:
        if "/" in glob_:
            directory, base = os.path.split(glob_)
            dir_ = here_ / directory
        else:
            dir_ = here_
            base = glob_
        match = [p for p in dir_.listdir() if p.fnmatch_basename(base)]
        matches.extend(match)
    result = [p for p in set(matches) if select(p)]
    return result if result[1:] else strings


def pyc_to_py(path_to_file):
    """Change some file extensions to those which are more likely to be text

    >>> pyc_to_py('vim.pyc') == 'vim.py'
    True
    """
    stem, ext = os.path.splitext(path_to_file)
    if ext == ".pyc":
        return f"{stem}.py"
    return path_to_file



"""Linux-specific code"""


from pysyte.types import paths


def xdg_home():
    """path to $XDG_CONFIG_HOME

    >>> assert xdg_home() == paths.path('~/.config').expand()
    """
    return paths.environ_path("XDG_CONFIG_HOME", "~/.config")


def xdg_home_config(filename):
    """path to that file in $XDG_CONFIG_HOME

    >>> assert xdg_home_config('fred') == paths.path('~/.config/fred').expand()
    """
    return xdg_home() / filename


def xdg_dirs():
    """paths in $XDG_CONFIG_DIRS"""
    return paths.environ_paths("XDG_CONFIG_DIRS")


def xdg_homes():
    return [xdg_home()]


bash_paste = "xclip -selection clipboard"
bash_copy = "xclip -selection clipboard -o"



"""Handle requirements for pysyte"""

from dataclasses import dataclass

from pysyte.types import versions
from pysyte.types.paths import dirs
from pysyte.types.paths import files
from pysyte.types.python import packages


@dataclass
class Requirement(packages.Package):
    """A python package required by something or other"""

    min_: versions.version = versions.NoVersion()
    max_: versions.version = versions.NoVersion()

    def __postinit__(self):
        import re

        breakpoint()
        if re.search('[<=>]', self.name):
            pass


@dataclass
class Requirements:
    """A collection of requirements"""

    required: list = [Requirement]

    def extend(self, items: list[Requirement]):
        self.required.extend(items)

    def append(self, item: Requirement):
        self.required.append(item)

    def pop(self) -> Requirement:
        return self.required.pop()

    def __iter__(self):
        yield from self.required


class RequirementDir(dirs.DirPath):
    """A directory, with requirements files as *.txt"""

    def requirement_files(self):
        """The requirements files in here"""
        return self.files('*.txt')


def parse(file: files.FilePath) -> Requirements:
    breakpoint()



"""Handle openai for pysyte"""

import openai

from pysyte.config import apis
from pysyte.ai import apis


@dataclass
class OpenaiApp:
    key_provider: str

    def __init__(self, key_provider):
        self.config = apis.ApiConfiguration(__file__, key_provider, "openai")

    def ask(self, messages):
        response = openai.ChatCompletion.create(
            messages=messages,
            model=self.config.model,
            n=self.config.times,
            max_tokens=self.config.tokens.max,
        )
        return response.choices


openai = JalanbConfiguration(__file__)
wwts = WwtsConfiguration()
app = OpenaiApp(wwts)



"""Handle arguments from the command line

This module is a simplifying proxy to stdlib's argparse
"""

import argparse
import os
import re
import shlex
import sys
from bdb import BdbQuit
from pprint import pformat
from functools import partial
from typing import Any
from typing import List

import stackprinter

from pysyte.cli.config import load_configs
from pysyte.types.numbers import inty

from pysyte.cli.config import pysyte


def config(arguments):
    return load_configs(arguments.prog)


def extract_strings(names: dict, name: str) -> List[str]:
    try:
        value = names[name]
    except KeyError:
        return []
    if isinstance(value, str):
        return [value]
    try:
        return [_ for _ in value if isinstance(_, str)]
    except TypeError:
        return []


class IntyAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        setattr(namespace, self.dest, inty(value))


class ArgumentsParser(object):
    """Add more attriubtes and methods to an argparse.ArgumentParser"""

    def __init__(self, argparser):
        self.parser = argparser
        self._parsed = None

        self.parse = self.parse_args

        self.string = self.arg = self.add_option
        self.boolean = self.opt = self.true = partial(self.arg, action="store_true")
        self.option = self.boolean
        self.integer = self.int = partial(self.arg, type=int)
        self.inty = partial(self.arg, action=IntyAction)
        self.strings = partial(self.positional, type=str)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def add_to(self, name=None, group=None, type_=None, default=None, *args, **kwargs):
        args_ = args[:]
        kwargs_ = kwargs.copy()
        name_ = name if name else ""
        group_ = group if group else name_
        kwargs_["default"] = (
            default if default else group_ if group_ else name_ if name_ else None
        )
        args_ = [name_[0], name_] + args
        self.add_option(*args_, **kwargs_)

    def add_option(self, initial, name, *args, **kwargs):
        n = name.lstrip("-")
        i = initial.lstrip("-") or n[0]
        return self.parser.add_argument(f"-{i}", f"--{n}", *args, **kwargs)

        if not initial:
            initial = name[0]
        return self.parser.add_argument(
            f'-{initial.lstrip("-")}', f'--{name.lstrip("-")}', *args, **kwargs
        )

    def positional(self, *args, **kwargs):
        """Add optional positional args"""
        return self.parser.add_argument(*args, **kwargs, nargs="*")

    def positionals(self, *args, **kwargs):
        """Add mandatory positional args"""
        return self.parser.add_argument(*args, **kwargs, nargs="+")

    def post_parser(self, args):
        """No-op for specialisation"""
        return args

    def parse_args(self, arguments=None, post_parser=None):
        post_parse = post_parser if post_parser else getattr(self, "post_parser")
        parsed_args = self.parser.parse_args(arguments)
        argument_namespace = ArgumentsNamespace(parsed_args)
        post_parser_ = post_parse if post_parse else lambda x: x
        posted_args = post_parser_(argument_namespace)
        self.args = posted_args
        self.args.prog = self.parser.prog
        return self.args

    def parse_string(self, string):
        return self.parse_args(shlex.split(string))


class DescribedParser(ArgumentsParser):
    """An argparse.ArgumentParser with a description, usage and epilog"""

    def __init__(self, description, usage, epilog):
        super().__init__(
            argparse.ArgumentParser(
                formatter_class=argparse.RawDescriptionHelpFormatter,
                description=description,
                usage=usage,
                epilog=epilog,
            )
        )


class ArgumentsNamespace(object):
    def __init__(self, result):
        self._result = result

    def __repr__(self):
        """Show args nicely in cli and debuggers"""
        klass = self.__class__.__qualname__
        cmd = " ".join(sys.argv)
        cmd_ = f"$ {cmd}"
        args = pformat(self.get_args())
        repr_ = "\n".join((klass, cmd_, args))
        return f"<{repr_}>"

    def __getattr__(self, name) -> Any:
        return getattr(self._result, name)

    def get_args(self) -> dict:
        attributes = [a for a in dir(self._result) if a[0] != "_"]
        return {a: getattr(self._result, a) for a in attributes}

    def get_arg(self, name: str) -> Any:
        return getattr(self._result, name, None)

    def get_strings(self, name: str) -> List[str]:
        if not self._result:
            return []
        return extract_strings(self._result.__dict__, name)

    def set_arg(self, name: str, value: Any) -> None:
        setattr(self._result, name, value)


class ArgumentHandler:
    def run(self, caller):
        """Handle expected exceptions"""
        try:
            return caller.main(self)
        except KeyboardInterrupt:
            sys.stderr.write("^c ^C ^c    ^C ^c ^C    ^c ^C ^c\n")
            ctrl_c = 3
            return ctrl_c
        except BdbQuit:
            return os.X_OK
        except SystemExit as e:
            return e.code
        except Exception as e:
            stackprinter.show(e, style=pysyte.stackprinter.style)


def parser(description=None, usage=None, epilog=None):
    """Make a command line argument parser"""

    if usage:
        return DescribedParser(description, usage, epilog)
    lines = (description or "").splitlines()
    regexp = re.compile("^[uU]sage:")
    usages = [_ for _ in lines if regexp.match(_)]
    if usages:
        usage_line = usages.pop()
        usage = usage_line.split(":", 1)[1]
        description = "\n".join([_ for _ in lines if _ != usage_line])
    elif len(lines) > 1 and not lines[1]:
        usage, description = lines[0], "\n".join(lines[2:])
    return DescribedParser(description, usage, epilog)


def test_parser():
    """A parser for testing convenience"""
    return parser("Testing", "Use this from a test", "")


def in_debugger():
    """Whether script is being traced

    A debugger will often set a system trace
    So this functions returns whether one has been set

    Debuggers known to give True
        * pudb
        * PyCharm

    Test runners known to give True
        * nose
        * py.test
        * coverage

    Test runners known to give False
        * doctest
        * unittest
    """
    return bool(sys.gettrace())



#! /usr/bin/env python3
"""Find imports in python files"""

import linecache

from pysyte import importers
from pysyte.cli.main import run
from pysyte.types import paths


def add_args(parser):
    """Parse out command line arguments"""
    parser.positional("source", help="path to source(s) to be checked")
    parser.boolean("", "edit", help="Show a command for editing")
    parser.boolean("", "multiple", help="Show multiple imports")
    parser.boolean("", "unused", help="Show unused imports")
    return parser


def texter(path):
    def text(line):
        string = "% 4d: %s" % (line, linecache.getline(path, line))
        return string.rstrip()

    return text


def show_unused(visitor):
    unused_lines = visitor.unused_lines()
    if not unused_lines:
        return []
    text = texter(visitor.path)
    print("Unused:")
    for line in sorted(unused_lines):
        names = unused_lines[line]
        print(",".join(names))
        print(text(line))
    return visitor.unused().keys()


def show_multiples(visitor):
    multiples = visitor.multiples()
    if multiples:
        print("Multiples:")
    for name, lines in multiples.items():
        instances = [visitor.line(_, True) for _ in lines]
        lines = [name] + instances
        print("\n".join(lines))
    return multiples.keys()


def find_sources(args):
    ignores = ["__pycache__", ".tox", ".git", ".venv"]
    result = []
    for arg in args:
        path_to_arg = paths.path(arg)
        if path_to_arg.isfile():
            result.append(path_to_arg)
        elif path_to_arg.isdir():
            for path_to_file in path_to_arg.walkfiles(pattern="*.py", ignores=ignores):
                result.append(path_to_file)
        else:
            raise TypeError(f"Unknown path type {path_to_arg}")
    return result


def show_imports(args, source):
    visitor = importers.parse(source)
    modules = []
    if args.multiple:
        modules.extend(show_multiples(visitor))
    if args.unused:
        modules.extend(show_unused(visitor))
    if args.edit and modules:
        sought = r"\|".join((rf"\<{_}\>" for _ in modules))
        print(f'\nvim {visitor.path} +/"{sought}"')
    return bool(modules)


def main(args) -> bool:
    result = False
    sources = find_sources(args.source)
    for source in sources:
        if show_imports(args, source):
            result = True
    return result


run(main, add_args)



import os
import atexit

from pysyte.bash.shell import run

_alt_screen_started = False

_TERM = "xterm"


def _stop_alt_screen():
    if _alt_screen_started:
        run(f"TERM={_TERM} tput rmcup")


def _start_alt_screen():
    global _alt_screen_started
    _alt_screen_started = True
    run("tput smcup")


def get_alt_screens():
    return _start_alt_screen, _stop_alt_screen


def alt_screen():
    global _TERM
    _TERM = os.environ.get("TERM", "xterm")
    atexit.register(_stop_alt_screen)
    return _start_alt_screen



"""Methods to provide X11 names for colours

If there is an "rgb.txt" file in a known directory get names from that
    otherwise fall back on the standard 8 names (BRGBCYMW)

The rgb.txt file is expected to contain X11 colour names
    so the "known directory" is <X11root>/lib/X11/
        Some variations on that official path are also tried
    See http://en.wikipedia.org/wiki/X11_color_names
"""


import os
import re


def _rgb_txt_directories():
    """List of directories which should contain rgb.txt

    Look in <X11root>/lib/X11/rgb.txt first
        then <X11root>/X11/share
        then in runtime directory
        then in same dir as this file
    """
    dirs = [
        "/usr/lib/X11",
        "/usr/X11/share/X11",
        ".",
        os.path.dirname(__file__),
    ]
    return dirs


def first_rgb_file(rgb_txt_paths):
    """The first rgb.txt file in known directories"""
    for path in rgb_txt_paths():
        path_to_txt_file = os.path.join(path, "rgb.txt")
        if os.path.isfile(path_to_txt_file):
            return path_to_txt_file
    return None


def _rgb_txt_line(string):
    """Parse a line from an X11 rgb.txt file

    Gives a name and 3 integers (RGB values)
    """
    regexp = re.compile(
        r"([ 0-9][ 0-9][ 0-9])\s+([ 0-9][ 0-9][ 0-9])\s+([ 0-9][ 0-9][ 0-9])"
        r"\s+([a-zA-Z0-9 ]+)\s*"
    )
    match = regexp.match(string)
    if not match:
        return "", (-1, -1, -1)
    red, green, blue, name = match.groups()
    return name.strip(), (int(red), int(green), int(blue))


def _rgb_txt_names_and_numbers(path_to_file):
    """Parse all lines from that file

    Expects that each line has a name and RGB values
    """
    if not path_to_file or not os.path.isfile(path_to_file):
        return []
    if not hasattr(_rgb_txt_names_and_numbers, "result"):
        rgb_lines = [_rgb_txt_line(_) for _ in open(path_to_file)]
        _rgb_txt_names_and_numbers.result = [
            (name, values) for name, values in rgb_lines if name
        ]
    return _rgb_txt_names_and_numbers.result


def _local_rgb_txt_names_and_numbers():
    """Names and RGB values from local machine"""
    path_to_rgb_txt = first_rgb_file(_rgb_txt_directories)
    return _rgb_txt_names_and_numbers(path_to_rgb_txt)


def names():
    """Dictionary of available colours as {name: (R, G, B)}

    Look for colours in an X11 rgb.txt

    >>> assert names()['green'] == (0, 255, 0)
    """
    names_and_numbers = _local_rgb_txt_names_and_numbers()
    return dict(names_and_numbers)



"""Methods for handling lines (of text)"""
import re


def _chop(lines_in, at, first, last):
    def as_int(string, start, end):
        try:
            return int(string)
        except (ValueError, TypeError):
            lines = lines_in[at:at] if isinstance(at, int) else lines_in[start:end]
            matcher = re.compile(string)
            for i, line in enumerate(lines, 1 + start):
                if matcher.search(line):
                    return i
        return 0

    def boundaries():
        def line(i):
            line = i if i >= 0 else length_read + 1 + i
            if line >= length_read:
                return length_read
            return 0 if line < 1 else line

        if at:
            start = first if first else 0
            end = last if last else -1
            first_ = line(as_int(at, start, end) - 1)
            return first_, first_ + 1
        first_ = line(as_int(first, 0, -1) - 1)
        last_ = line(as_int(last, first_, -1) or -1)
        return first_, last_

    length_read = len(lines_in)
    first, last = boundaries()
    rest = [] if first > length_read else lines_in[first:last]
    return first, rest


def chop(text, at, first=1, last=-1):
    r"""Make that text fit those boundaries

    >>> lines = ['one', 'two', 'three', 'four']
    >>> text = '\n'.join(lines)

    >>> assert 'three' not in chop(text, 1, last=2)

    >>> assert chop(text, at=2) == chop(text, '[t][w][o]')
    """
    return _chop(text.splitlines() or [], at, first, last)


def set_width(line, width):
    if not width:
        return line
    return line[:width]


def _number_format(count=999):
    """A string format for line numbers

    Should give a '%d' format with width big enough to `count` lines

    >>> assert _number_format(77) == '%2d: '
    """
    digits = len(str(count))
    return "%%%dd: " % digits


def add_numbers(lines, first=0):
    def numbered(line_, line_format_):
        prefix = line_format_ % (first + i + 1)
        return f"{prefix}{line_.rstrip()}"

    for i, line in enumerate(lines):
        yield numbered(line, _number_format(len(lines) + first))


def sed(lines, args):
    return reformat_lines(
        lines, first=args.first, numbers=args.numbers, width=args.width
    )


def reformat_lines(lines, first, numbers, width):
    adder = add_numbers if numbers else lambda x, y: x
    return list(adder([_[:width] if width else _ for _ in lines], first))


def select(predicate, lines):
    for line in lines:
        if not predicate(line):
            continue
        yield line


_selector = lambda method, lines_: (_ for _ in lines_ if method(_))
_converter = lambda method, lines_: (method(_) for _ in lines_)

full = lambda lines_: _converter(
    lambda line: line.rstrip(), _selector(lambda line: line.rstrip(), lines_)
)
grep = lambda lines_: lambda regexp: _selector(
    lambda line: line and re.search(regexp, line), lines_
)


def arg_lines(text, args):
    return chop(text, args.at, args.first, args.last)


def as_text(lines):
    return "\n".join(lines)



import inspect
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable
from typing import Optional
from types import FrameType
from types import ModuleType


def unwrap(method: Callable) -> Callable:
    """Get the original method from a methodeven if it's wrapped """
    return getattr(method, "__wrapped__", method)


@dataclass
class MethodData:
    method: Callable


class Method(MethodData):
    """A callable method with some convenience attributes"""

    def __init__(self, method: Callable):
        super().__init__(unwrap(method))
        if self.method != method:
            self.wrapped = method
        self.code = self.method.__code__
        self.init_frame = inspect.currentframe()
        self.callers = [self.init_frame.f_back]

    def run(self, *args, **kwargs):
        return self.method(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        call_frame = inspect.currentframe()
        self.callers.append(call_frame.f_back)
        return self.method(*args, **kwargs)

    @property
    def caller(self) -> Optional[FrameType]:
        return self.callers[-1]

    @property
    def filename(self) -> str:
        return self.code.co_filename

    @property
    def module(self) -> Optional[ModuleType]:
        return inspect.getmodule(self.method)

    @property
    def doc(self) -> str:
        return inspect.getdoc(self.method) or ""

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            if name == "__file__":
                name = "filename"
            if hasattr(self.code, f"co_{name}"):
                return getattr(self.code, f"co_{name}")
            raise


@contextmanager
def caller():
    yield inspect.currentframe().f_back.f_back.f_back


def _represent_args(*args, **kwargs):
    """Represent the aruments in a form suitable as a key (hashable)

    And which will be recognisable to user in error messages
    >>> print(_represent_args([1, 2], **{'fred': 'here'}))
    [1, 2], fred='here'
    """
    argument_strings = [repr(a) for a in args]
    keyword_strings = ["=".join((k, repr(v))) for k, v in kwargs.items()]
    return ", ".join(argument_strings + keyword_strings)


def memoized(method):
    """A new method which acts like the given method but memoizeds args

    See https://en.wikipedia.org/wiki/Memoization for the general idea
        >>> @memoized
        ... def test(arg):
        ...     print('called')
        ...     return arg + 1
        ...
        >>> test(1)
        called
        2
        >>> test(2)
        called
        3
        >>> test(1)
        2

    The returned method also has an attached method "invalidate"
        which removes given values from the cache
        Or empties the cache if no values are given
        >>> test.invalidate(2)
        >>> test(1)
        2
        >>> test(2)
        called
        3
    """
    method.cache = {}

    def invalidate(*args, **kwargs):
        key = _represent_args(*args, **kwargs)
        if not key:
            method.cache = {}
        elif key in method.cache:
            del method.cache[key]
        else:
            raise KeyError(f"Not prevously cached: {method.__name__}({key})")

    def new_method(*args, **kwargs):
        """Cache the args and return values of the call

        The key cached is the repr() of args
            This allows more types of values to be used as keys to the cache
            Such as lists and tuples
        """
        key = _represent_args(*args, **kwargs)
        if key not in method.cache:
            method.cache[key] = method(*args, **kwargs)
        return method.cache[key]

    new_method.invalidate = invalidate
    new_method.__doc__ = method.__doc__
    new_method.__name__ = f"memoized({method.__name__})"
    return new_method



"""Module to handle streams of text from cli arguments"""

import os
import sys

from six import StringIO

from pysyte import iteration
from pysyte.cli import arguments
from pysyte.oss.platforms import get_clipboard_data


def parse_args(description=""):
    """Parse out command line arguments"""
    parser = arguments.parser(description or __doc__)
    parser.positional("streams", help="streams to use")
    parser.boolean("p", "paste", help="paste text from clipboard")
    parser.boolean("i", "stdin", help="wait for text from stdin")
    return parser.parse_args()


def args(parsed_args, name="streams", files_only=False):
    """Interpret parsed args to streams"""
    streams = []
    strings = parsed_args.get_strings(name)
    files = [s for s in strings if os.path.isfile(s)]
    if files:
        streams = [open(f) for f in files]
    if strings or files_only:
        return streams
    if "-" in files or getattr(parsed_args, "stdin", False):
        streams = [sys.stdin]
    if getattr(parsed_args, "paste", False):
        streams.append(clipboard_stream())
    return streams


def files(parsed_args, name=None):
    return args(parsed_args, name, True)


def all():
    yielded = False
    for path in _arg_files():
        with open(path) as stream:
            yield stream
        yielded = True
    if not yielded or "-" in sys.argv:
        yield sys.stdin


def some():
    if sys.argv[1:]:
        assert _arg_files()
    return any()


def clipboard_stream(name=None):
    stream = StringIO(get_clipboard_data())
    stream.name = name or "<clipboard>"
    return stream


def _arg_files():
    return [a for a in sys.argv[1:] if os.path.isfile(a)]


def _arg_streams():
    """yield streams to all arg.isfile()"""
    for path in _arg_files():
        with open(path) as stream:
            yield stream


def any():
    try:
        stream = iteration.first(_arg_streams())
        if stream:
            return _arg_streams()
    except ValueError:
        return iter([clipboard_stream(), sys.stdin])



"""Provide a runnable app to pysyte"""

from bdb import BdbQuit
import sys

from contextlib import ContextDecorator

from pysyte.types.methods import Method
from pysyte.cli.exceptions import rich_exceptions
from pysyte.cli import exits


def exit(method, locals_=None):
    exit_code = exits.fail
    with rich_exceptions(locals_ if locals_ else {}):
        try:
            exit_code = exits.ExitCode(method())
        except BdbQuit:
            exit_code = exits.pass_
    sys.exit(exit_code)


class App(Method, ContextDecorator):
    def __enter__(self):
        self.exit_code = None
        return self

    def run(self, *args, **kwargs):
        kwargs["app"] = self
        self.exit_code = super().run(*args, **kwargs)

    def __exit__(self, *exc):
        return self.exit_code



#! /usr/bin/env python3
"""Script to get a key from the keyboard"""

from __future__ import print_function


from pysyte.oss import getch
from pysyte.cli.main import run


__version__ = "0.1.1"


class ScriptError(NotImplementedError):
    pass


def add_args(parser):
    parser.boolean("", "codes", help="Show raw codes")
    parser.boolean(
        "",
        "prompt",
        default="",
        help="Prompt to show before getting keys " "(default: none)",
    )
    parser.boolean("", "string", help="Show string")
    return parser


def main(args):
    if args.codes:
        print(", ".join(getch.get_codes()))
    elif args.string:
        print(getch.get_string())
    else:
        print(getch.get_key())
    return True


run(main, add_args)



"""Some methods to help with the handling of dictionaries"""

from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from yamlreader import data_merge


def get_caselessly(dictionary, sought):
    """Find the sought key in the given dictionary regardless of case

    >>> things = {'Fred': 9}
    >>> print(get_caselessly(things, 'fred'))
    9
    """
    try:
        return dictionary[sought]
    except KeyError:
        caseless_keys = {k.lower(): k for k in dictionary.keys()}
        real_key = caseless_keys[sought.lower()]  # allow any KeyError here
        return dictionary[real_key]


def swap_dictionary(dictionary):
    """Swap keys for values in the given dictionary

    >>> swap_dictionary({'one': 1})[1] == 'one'
    True
    """
    if dictionary is None:
        return None
    return {v: k for k, v in dictionary.items()}


def append_value(dictionary, key, item):
    """Append those items to the values for that key"""
    items = dictionary.get(key, [])
    items.append(item)
    dictionary[key] = items


def extend_values(dictionary, key, items):
    """Extend the values for that key with the items"""
    values = dictionary.get(key, [])
    try:
        values.extend(items)
    except TypeError:
        raise TypeError("Expected a list, got: %r" % items)
    dictionary[key] = values


def group_list(items):
    """Items in the list are grouped into a dictionary

    Items should be a list of (key, value) pairs

    >>> grouped = group_list([(1, 0), (2, 0), (1, 1)])
    >>> grouped[1] == [0, 1]
    True
    """
    groups = defaultdict(list)
    for key, value in items:
        groups[key].append(value)
    return groups


def group_list_by(items, key_from_item):
    """Items in the list are grouped into a dictionary

    key_from_item is a method to get the key from the item

    >>> items = [(1, 9), (2, 8), (1, 7)]
    >>> key_from_item = lambda x: 'a' if x[0] == 1 else 'b'
    >>> grouped = group_list_by(items, key_from_item)
    >>> grouped['a'] == [(1, 9), (1, 7)]
    True
    """
    groups = defaultdict(list)
    for item in items:
        key = key_from_item(item)
        groups[key].append(item)
    return groups


class DefaultDict(defaultdict):
    """A default dict with improved repr"""

    def __repr__(self):
        from pprint import pformat

        name = self.__class__.__name__
        value = pformat(dict(self), indent=4)
        return f"<{name}\n    {value}\n>"


class LazyDefaultDict(DefaultDict):
    """A defaultdict which waits for key access to provide default

    Same as a defaultdict, but the initiasing method takes a key
        and provides a value
    """

    def __init__(self, method):
        self.method = method
        super(LazyDefaultDict, self).__init__(None)

    def __missing__(self, key):
        result = self.method(key)
        self[key] = result
        return result


class NameSpace(dict):
    """Convert dictionary keys to attributes of self

    >>> assert NameSpace({'fred': 1}).fred == 1
    """

    def __init__(self, *args, **kwargs):
        """>>> assert NameSpace({'fred': 1}).fred == 1"""
        super(NameSpace, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def update(self, other):
        self.__dict__ = data_merge(self.__dict__, other)


class NameSpaces(NameSpace):
    """Convert dictionary keys to attributes of self, recursively

    >>> instance = NameSpaces({'fred': {'mary': 1}})
    >>> assert instance.fred.mary == 1
    """

    def __init__(self, thing):
        data = {}
        for key, value in (thing or {}).items():
            data[key] = NameSpaces(value) if isinstance(value, dict) else value
        super(NameSpaces, self).__init__(data)


@dataclass
class SpaceNameData:
    proxy: Any


class SpaceName(SpaceNameData):
    """Access attributes of a thing like a dict"""

    def __getitem__(self, name):
        return self.getitem(name)

    def getitem(self, name):
        try:
            return getattr(self.proxy, name)
        except AttributeError:
            raise KeyError(name)


class SpaceNames(SpaceName):
    def getitem(self, name_):
        name, *names_ = name_.split(".", 1)
        value = super().getitem(name)
        try:
            return SpaceNames(value).getitem(".".join(names))
        except IndexError:
            return value



"""Example usage of OpenAPI calls"""

import sys

from rich import print

from pysyte.cli import exits
from pysyte.ai.open import OpenaiApp
from pysyte.oss.getch import ask_user_simplified
from pysyte.types.dictionaries import NameSpaces


models = NameSpaces(
    dict(
        davinci="text-davinci-003",
        curie="text-curie-001",
        babbage="text-babbage-001",
        ada="text-ada-001",
    )
)


def main() -> exits.ExitCode:

    app = OpenaiApp("wwts")

    breakpoint()
    question = " ".join(sys.argv[1:]) or config.prompt.final
    messages = [
        {"role": "system", "content": config.prompt.prefix},
        {"role": "system", "content": config.prompt.context},
        {"role": "system", "content": config.prompt.task},
        {"role": "system", "content": config.prompt.rules},
        {"role": "user", "content": question},
    ]
    choices = app.ask([], config)

    allowed = []
    for i, choice in enumerate(choices, 1):
        reply = choice['message']['content']
        print(f"Reply {i}: {reply}\n\n")
        allowed.append(str(i))
    answer = ask_user_simplified("Which reply is best?", "")
    if answer not in allowed:
        return exits.fail
    choice = choices[int(answer) - 1]
    messages.append({"role": "assistant", "content": choice['message']['content']})
    question = input("Ask more? ")
    if not question:
        return exits.pass_
    messages.append({"role": "user", "content": question})
    choices = app.ask(messages, config)
    for i, choice in enumerate(choices, 1):
        reply = choice['message']['content']
        print(f"Reply {i}: {reply}\n\n")
    return exits.pass_


if __name__ == "__main__":
    sys.exit(int(main()))



"""Handle bash commands for pysyte"""

import os
from contextlib import contextmanager
from subprocess import getstatusoutput
from typing import List

from boltons.setutils import IndexedSet


class BashError(ValueError):
    pass


_working_dirs: List[str] = [""]
_paths: List[str] = []


def cd(path: str):
    if _working_dirs[0] == path:
        return
    assert os.path.isdir(path)
    _working_dirs[0] = path


@contextmanager
def pushd(path):
    assert os.path.isdir(path)
    _working_dirs.insert(0, path)
    yield
    del _working_dirs[0]


def _get_path() -> str:
    """Guarantee that /usr/local/bin, /usr/bin, /bin are in PATH"""
    if _paths:
        return _paths[0]
    environ_paths = IndexedSet(os.environ["PATH"].split(":"))
    minimal_paths = IndexedSet(["/usr/local/bin", "/usr/bin", "/bin"])
    all_paths = environ_paths | minimal_paths
    _paths.append(":".join(all_paths))
    return _paths[0]


def run(command: str) -> str:
    path_command = f"PATH={_get_path()} {command}"
    if _working_dirs[0]:
        run_command = f"(cd {_working_dirs[0]}; {path_command})"
    else:
        run_command = path_command
    status, output = getstatusoutput(run_command)
    if status:
        raise BashError(f"{run_command}\n{output}")
    return output



import os
import setuptools


def package_files(directory):
    paths = []
    extensions = (".test", ".tests", ".yaml")
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            extension = os.path.splitext(filename)[-1]
            if extension not in extensions:
                continue
            paths.append(os.path.join("..", path, filename))
    return paths


setuptools.setup(
    packages=setuptools.find_packages(),
    package_data={"": package_files("pysyte")},
    install_requires=[
        "boltons",
        "deprecated",
        "inflect>=2.1.0",
        "path.py==7.7.1",
        "pyyaml",
        "rich",
        "stackprinter",
        "yamlreader",
    ],
)



"""Some keyboard handling code"""


import sys


from pysyte.oss import getch


def _get_chosen_chars(chooser):
    while True:
        key = getch.get_as_key()
        try:
            if chooser(key):
                return key
        except AttributeError:
            print(key)
            continue


def get_digit():
    return _get_chosen_chars(lambda x: x.isdigit())


def get_letter():
    return _get_chosen_chars(lambda x: x.isupper() or x.islower())


def quit_on_q():
    try:
        key = getch.get_as_key()
        if key in "qQ":
            sys.exit()
        return key
    except KeyboardInterrupt:
        sys.exit()



"""Handle numbers for pysyte

>>> from pysyte.types.literals import numbers

>>> assert numbers.ten == 10
>>> assert numbers.twelve == 12
>>> assert numbers.ninety_nine == 99

"""
from bidict import bidict

zero = 0


def on_import():
    digits = (
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
    )
    teens = (
        "eleven",
        "twelve",
        "thirteen",
        "fourteen",
        "fifteen",
        "sixteen",
        "seventeen",
        "eighteen",
        "nineteen",
    )
    tens = (
        "ten",
        "twenty",
        "thirty",
        "forty",
        "fifty",
        "sixty",
        "seventy",
        "eighty",
        "ninety",
    )

    data = {}
    data.update({v: k for k, v in enumerate(digits, 1)})
    data.update({v: k for k, v in enumerate(teens, 11)})
    for i, dd in enumerate(tens, 1):
        data[dd] = i * 10
        if i == 1:
            continue
        for j, d in enumerate(digits, 1):
            data[f"{dd}_{d}"] = i * 10 + j
    return data


def name(n: int) -> str:
    """Get the name of an integer

    >>> assert name(0) == "zero"
    >>> assert name(7) == "seven"
    >>> assert name(123) == "one hundred and twenty three"
    >>> assert name(2_000) == "two thousand"
    >>> assert name(2_001) == "two thousand and one"
    >>> assert name(2_100) == "two thousand, one hundred"
    >>> assert name(2_101) == "two thousand, one hundred and one"
    >>> assert name(54321) == "fifty four thousand, three hundred and twenty one"
    >>> assert name(2_000_000) == "two million"
    >>> assert name(2_000_001) == "two million and one"
    >>> assert name(2_000_100) == "two million and one hundred"
    >>> assert name(2_001_000) == "two million and one thousand"
    >>> assert name(2_000_101) == "two million, one hundred and one"
    >>> assert name(2_001_001) == "two million, one thousand and one"
    >>> assert name(2_001_100) == "two million and one thousand, one hundred"
    >>> assert name(2_001_101) == "two million, one thousand, one hundred and one"
    >>> assert (
    ...     name(7654321)
    ...     == "seven million, six hundred and fifty four thousand, three hundred and twenty one"
    ... )
    >>> assert (
    ...     name(77654021)
    ...     == "seventy seven million, six hundred and fifty four thousand and twenty one"
    ... )
    >>> assert (
    ...     name(7777654001)
    ...     == "seven billion, seven hundred and seventy seven million, six hundred and fifty four thousand and one"
    ... )
    >>> assert (
    ...     name(12000000000000000000007890000000000000000000045391)
    ...     == "twelve quindecillion, seven octillion, eight hundred and ninety heptillion, forty five thousand, three hundred and ninety one"
    ... )
    """

    def bigger_name(n: int, m: int, units: str) -> str:
        """Get the name of a bigger number

        This will only be called with n > 1_000_000
        """
        assert n > m
        div, mod = divmod(n, m)
        assert div
        divs_ = f"{name(div)} {units}" if 0 < div else ""
        if not mod:
            return divs_
        mods_ = name(mod)
        and_ = ", " if " and " in mods_ else " and "
        result = f"{divs_}{and_}{mods_}"
        return result

    if n < 0:
        return f'minus {name(-n)}'
    result_ = f"{n:_}"
    data = {k: v for k, v in globals().items() if isinstance(v, int)}
    numbers = dict(bidict(data).inverse)
    if n in numbers:
        return numbers[n].replace('_', ' ')
    thousands, units = divmod(n, 1_000)
    hundreds, tens = divmod(units, 100)
    tens_ = f"{name(tens)}" if tens else ""
    hundreds_ = f"{name(hundreds)} hundred" if 0 < hundreds < 100 else ""
    and_ = " and " if hundreds and tens else ""
    result = f'{hundreds_}{and_}{tens_}'
    if n < 1_000:
        return result
    and_ = ""
    if thousands:
        if tens:
            and_ = " and "
        if hundreds:
            and_ = ", "
    thousands_ = f"{name(thousands)} thousand" if 0 < thousands else ""
    result = f"{thousands_}{and_}{result}"
    if n < 1_000_000:
        return result
    prefixes = (
        "m",
        "b",
        "tr",
        "quadr",
        "quint",
        "sext",
        "hept",
        "oct",
        "non",
        "dec",
        "undec",
        "duodec",
        "tredec",
        "quattuordec",
        "quindec",
        "sexdec",
        "septdec",
        "octodec",
        "novemdec",
        "vigint",
    )
    i = 1_000_000
    for prefix in prefixes:
        j = i * 1_000
        if n < j:
            name_ = f"{prefix}illion"
            return bigger_name(n, i, name_)
        i = j
    return "bigger"


globals().update(on_import())


del on_import

if __name__ == '__main__':
    import doctest

    doctest.testmod()



from dataclasses import dataclass

from pysyte.config.types import ModuleConfiguration


@dataclass
class Apis:
    """A provider of APIs"""

    name: str = "Fred"


@dataclass
class ApiConfiguration(ModuleConfiguration):
    api: Apis

    def __init__(file: str, key: str, api: Apis):
        super().__init__(file)
        openai.api_key = self.keys[key][api.name]



"""Import imports for pysyte"""
import abc
import os
import importlib
import linecache
import ast
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator

from pym.ast.visit import visitors


class ImportVisitor(visitors.PymVisitor):
    def __init__(self):
        super().__init__()
        self.imports = defaultdict(list)
        self.froms = defaultdict(list)

    @abc.abstractmethod
    def imported(self, name, line):
        pass

    def collect_names(self, node):
        names = [(_.name, getattr(_, "asname", "")) for _ in node.names]
        for name, alias in names:
            self.imports[alias if alias else name].append(node.lineno)
        return names

    def find_value_id(self, node, attr=None):
        if not node:
            return None
        value = getattr(node, "value", None)
        if not value:
            if attr:
                value = getattr(getattr(node, attr, None), "value", None)
                if not value:
                    name = getattr(getattr(node, attr, None), "id", None)
                    if name:
                        return name
        if not value:
            return getattr(node, "id", None)
        result = getattr(value, "id", None)
        if result:
            return result
        return self.find_value_id(value, attr)

    def find_name(self, node, *args):
        name = getattr(node, "id", None)
        if name is not None:
            return name
        for attribute in args:
            attr = getattr(node, attribute, None)
            if attr:
                name = self.find_name(attr, *args)
                if name:
                    return name
        return None

    def visit_ImportFrom(self, node):
        if node.module != "__future__":
            names = self.collect_names(node)
            self.froms[node.module].extend(names)
        self.generic_visit(node)

    def visit_Import(self, node):
        _ = self.collect_names(node)
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        for decorator in node.decorator_list:
            name = self.find_name(decorator, "func", "value")
            self.imported(name, decorator)
        self.generic_visit(node)

    def visit_Subscript(self, node):
        name = self.find_value_id(node)
        name2 = self.find_name(node, "value")
        assert name == name2
        self.imported(name, node)
        subname = self.find_value_id(node.slice)
        assert subname == self.find_name(node.slice, "value")
        self.imported(subname, node)
        self.generic_visit(node)

    def visit_Attribute(self, node):
        try:
            self.imported(node.value.id, node.lineno)
            full_name = f"{node.value.id}.{node.attr}"
            self.imported(full_name, node.lineno)
        except AttributeError:
            pass
        self.generic_visit(node)

    def visit_ClassDef(self, node):
        for base in node.bases:
            name = self.find_name(base, "value")
            self.imported(name, node)
        self.generic_visit(node)

    def visit_Call(self, node):
        func = node.func
        try:
            name = func.id
        except AttributeError:
            name = self.find_value_id(func, "func")
            name2 = self.find_name(node, "value", "func")
            assert name == name2
        self.imported(name, node)
        self.generic_visit(node)

    def visit_Name(self, node):
        name = self.find_name(node)
        self.imported(name, node)
        self.generic_visit(node)


class UsedImportVistor(ImportVisitor):
    def __init__(self):
        super().__init__()
        self.used = defaultdict(list)

    def imported(self, name, node):
        if not name:
            return
        if name in self.imports:
            self.used[name].append(node.lineno)

    def unused(self):
        return {k: v for k, v in self.imports.items() if k not in self.used}

    def unused_lines(self):
        result = defaultdict(set)
        for name, lines in self.unused().items():
            for line in lines:
                result[line].add(name)
        return result

    def multiples(self):
        return {k: v for k, v in self.imports.items() if len(v) > 1}

    def line(self, line_number, with_number=True):
        line = linecache.getline(self.path, line_number).rstrip()
        if not with_number:
            return line
        return f"{line_number:4d}: {line}"


@dataclass
class AS3:
    path: str
    tree: ast.Module


def find_imports(as3: AS3) -> UsedImportVistor:
    import_user = UsedImportVistor()
    import_user.visit(as3.tree)
    return import_user


@contextmanager
def parse_python(script) -> Iterator[AS3]:
    with open(script) as stream:
        as3 = ast.parse(stream.read(), script)
        yield AS3(script, as3)


def parse(script) -> UsedImportVistor:
    """Extract all imports from a python script"""
    if not os.path.isfile(script):
        raise FileNotFoundError(f"Not a file: {script}")
    with parse_python(script) as as3:
        importer = find_imports(as3)
        importer.path = script  # type: ignore [attr-defined]
        return importer


@contextmanager
def importer(module):
    """Provide a context with that module

    >>> with importer(os) as os_, importer('pysyte.importers') as importers:
    ...     assert os_.path is os.path
    ...     assert importer.__code__ == importers.importer.__code__
    ...
    """
    if isinstance(module, type(os)):
        yield module
    elif isinstance(module, str):
        yield importlib.import_module(module)



"""Handle iterators for pysyte"""

from typing import Sequence
from typing import TypeVar

T = TypeVar("T")  # Declare Type variable


def first(sequence: Sequence[T], message=None) -> T:  # Generic function
    """The first item in that sequence

    If there aren't any, raise a ValueError with that message

    >>> assert first([1, 2, 3]) == 1
    """
    try:
        return next(iter(sequence))
    except StopIteration:
        raise ValueError(message or (f"Sequence is empty: {sequence}"))


def last(sequence: Sequence[T], message=None) -> T:
    """The last item in that sequence

    If there aren't any, raise a ValueError with that message

    >>> assert last([1, 2, 3]) == 3
    """
    return first(list(reversed(sequence)), message)


def first_or(sequence: Sequence[T], value) -> T:
    """First item in that sequence, or that value

    >>> assert first_or([1, 2, 3], 4) == 1
    >>> assert first_or([], 4) == 4
    """
    try:
        return first(sequence)
    except ValueError:
        return value


def first_that(predicate, sequence: Sequence[T], message=None) -> T:
    """The first item in that sequence that matches that predicate

    If none matches raise a KeyError with that message

    >>> assert first_that(lambda x: x > 1, [1, 2, 3]) == 2
    """
    try:
        return first([_ for _ in sequence if predicate(_)])
    except (ValueError, StopIteration):
        raise KeyError(message or "Not Found")



"""Handle different OS platforms"""
import platform as python_platform
import importlib
from subprocess import run

name = python_platform.system().lower()
platform = importlib.import_module(f"pysyte.oss.{name}", "pysyte.oss")


def put_clipboard_data(data):
    run(platform.bash_copy, encoding="utf-8", input=data)


def get_clipboard_data():
    result = run(platform.bash_paste, capture_output=True, encoding="utf-8")
    return result.stdout
